"""
Provides function(s) to perform normalization techniques on CFG's
"""

import copy
import numpy as np
from ..utils import progressbar, update_memcfg_tokens, AtomicTokenDict
from .norm_utils import get_normalizer
from ..utils.type_utils import *
import bincfg


def normalize_cfg_data(cfg_data: 'Union[CFGInputDataType, bincfg.CFG, bincfg.MemCFG, bincfg.CFGDataset, bincfg.MemCFGDataset, Iterable]', 
                       normalizer: 'Union[str, NormalizerType]', inplace: 'bool' = False, using_tokens: 'Optional[Union[dict[str, int], AtomicTokenDict]]' = None, 
                       force_renormalize: 'bool' = False, convert_to_mem: 'bool' = False, conv_keep_mem_addrs: 'bool' = True, 
                       unpack_cfgs: 'bool' = False, progress: 'bool' = False) -> 'Union[bincfg.CFG, bincfg.MemCFG, bincfg.CFGDataset, bincfg.MemCFGDataset, list, tuple]':
    """Normalizes some cfg data.

    Args:
        cfg_data (Union[CFGInputDataType, CFG, MemCFG, CFGDataset, MemCFGDataset, Iterable]): some cfg data. Can be either: str, CFG, MemCFG, 
            CFGDataset, MemCFGDataset, or iterable of previously mentioned types. Will return the same type as that passed,
            unless that particular input was a string, in which case a CFG will be returned.
        normalizer (Union[str, Normalizer]): the normalizer to use. Can be either a ``Normalizer`` class with a 
            `.normalize()` method, or a string to use a built-in normalizer. See :func:`bincfg.normalization.get_normalizer`
            for acceptable strings.
        inplace (bool): if True, will modify data in-place instead of creating new objects. Defaults to False.
            NOTE: if `inplace=False`, and the incoming data has already been normalized with the passed `normalizer`, then
            the original cfg will be returned, NOT a copy.
        using_tokens (Optional[Union[dict[str, int], AtomicTokenDict]]): only used for ``MemCFG``'s. If not None, then a dictionary mapping string
            tokens to integer token values that will be used as any ``MemCFG``'s tokens. Defaults to None.
        force_renormalize (bool): by default, this method will only normalize cfg's whose .normalizer != to the passed
            normalizer. However if `force_renormalize=True`, then all cfg's will be renormalized even if they have been
            previously normalized with the same normalizer. Defaults to False.
        convert_to_mem (bool): if True, will convert all ``CFG``'s and ``CFGDatasets`` to their memory-efficient 
            versions after normalizing. Defaults to False.
        conv_keep_mem_addrs (bool): if True, will pass `keep_memory_addresses=True` when converting CFG's into MemCFG's
        unpack_cfgs (bool): by default, this method will return the same types that were passed to be normalized. 
            However if `unpack_cfgs=True`, then instead, a list of all cfgs unpacked (EG: unpacked from lists, and pulled 
            out of datasets) will be returned. Defaults to False.
            NOTE: if only a single ``CFG``/``MemCFG`` was passed, a list will still be returned of only that single element.
        progress (bool): if True, will show a progressbar for normalizations of multiple cfg's. Defaults to False.

    Returns:
        Union[CFG, MemCFG, CFGDataset, MemCFGDataset, List, Tuple]: the normalized data
    """
    normalizer = get_normalizer(normalizer)

    # If input is not a str/CFG/MemCFG/dataset, then it should be an iterable, loop or use multiprocessing
    if not isinstance(cfg_data, (str, bincfg.CFG, bincfg.MemCFG, bincfg.CFGDataset, bincfg.MemCFGDataset)):
        cfg_data = list(cfg_data)  # Copy the element references so we don't modify the original iterable, and convert to list
        cfgs, cfg_parts = _unpack_cfgs(cfg_data)

        # Normalize all the cfg's
        for i, cfg in enumerate(progressbar(cfgs, progress=progress)):
            cfgs[i] = normalize_cfg_data(cfg, normalizer=normalizer, inplace=inplace, using_tokens=using_tokens,
                force_renormalize=force_renormalize, convert_to_mem=convert_to_mem, progress=False)
        
        # Now that all our cfg's have been normalized, sort them back to where they belong (if needed)
        if not unpack_cfgs:
            for i, (start, end) in enumerate(cfg_parts):

                # Append the data if it is just a cfg, otherwise make it into a dataset
                if isinstance(cfg_data[i], (bincfg.CFG, bincfg.MemCFG)):
                    cfg_data[i] = cfgs[start]
                
                # Otherwise make it into a dataset
                else:
                    if isinstance(cfg_data[i], bincfg.CFGDataset) and convert_to_mem:
                        cfg_data[i] = bincfg.MemCFGDataset()
                    elif not inplace:  # Create a new object if not inplace
                        cfg_data[i] = cfg_data[i].__class__()

                    cfg_data[i].cfgs = cfgs[start:end]
                    cfg_data[i].normalizer = normalizer
            return cfg_data

        return cfgs
    
    # Check for a string and convert to a CFG right away. We can turn on inplace if so
    if isinstance(cfg_data, str):
        cfg_data = bincfg.CFG(cfg_data)
        inplace = True

    # Check if we even need to normalize
    if cfg_data.normalizer is not None and normalizer == cfg_data.normalizer and not force_renormalize:
        # Check if we need to unpack cfgs, and/or convert to memcfg's
        if convert_to_mem:
            if isinstance(cfg_data, bincfg.CFG):
                cfg_data = bincfg.MemCFG(cfg_data, using_tokens=using_tokens, keep_memory_addresses=conv_keep_mem_addrs)
            elif isinstance(cfg_data, bincfg.CFGDataset):
                cfg_data = bincfg.MemCFGDataset(cfg_data, using_tokens=using_tokens, keep_memory_addresses=conv_keep_mem_addrs)
            
            # Otherwise it is already a MemCFG/MemCFGDataset, and we need to check if we should update the tokens
            elif using_tokens is not None:
                update_memcfg_tokens(cfg_data, using_tokens)
        
        # Otherwise, check if this is already a memcfg/memcfgdataset and we need to update them tokens
        elif using_tokens is not None:
            if isinstance(cfg_data, (bincfg.MemCFG, bincfg.MemCFGDataset)):
                update_memcfg_tokens(cfg_data, using_tokens)

        if unpack_cfgs:
            if isinstance(cfg_data, (bincfg.CFGDataset, bincfg.MemCFGDataset)):
                cfg_data = cfg_data.cfgs
            else:
                cfg_data = [cfg_data]

        return cfg_data

    # Create return object depending on input type and inplace, set the return object's normalizer
    ret = cfg_data if inplace else copy.deepcopy(cfg_data) if isinstance(cfg_data, (bincfg.CFG, bincfg.MemCFG)) else cfg_data.__class__()
    ret.normalizer = normalizer
    
    # Single CFG's and MemCFG's will be normalized here
    if isinstance(cfg_data, bincfg.CFG):
        for block in ret.blocks:
            block.asm_lines = normalizer.normalize(*block.asm_lines, cfg=ret, block=block)

        # Convert to MemCFG if needed
        if convert_to_mem and isinstance(cfg_data, bincfg.CFG):
            ret = bincfg.MemCFG(ret, inplace=True, using_tokens=using_tokens, keep_memory_addresses=conv_keep_mem_addrs)

    elif isinstance(cfg_data, bincfg.MemCFG):
        # Need to recompute the asm_lines, block_asm_idx, and tokens
        # Need to start new_block_asm_idx with a 0
        norm_lines, new_block_asm_idx = [], [0]
        inv_tokens = {v: k for k, v in ret.tokens.items()}

        # Go through each block re-normalizing its lines, and keeping track of the block_asm_idx
        for block_idx in range(ret.num_blocks):
            norm_lines += normalizer.normalize(*[inv_tokens[t] for t in ret.get_block_asm_lines(block_idx)], cfg=ret, block=block_idx)
            new_block_asm_idx.append(len(norm_lines))
        
        # Update the new tokens if needed, allowing for use of atomic tokens
        if isinstance(using_tokens, AtomicTokenDict):
            using_tokens.addtokens(*norm_lines)
            new_tokens = using_tokens

        else:
            new_tokens = {} if using_tokens is None else using_tokens
            for t in norm_lines:
                new_tokens.setdefault(t, len(new_tokens))
        
        # Convert to integer tokens
        new_asm_lines = [new_tokens[t] for t in norm_lines]

        # Set the new data in ret
        ret.asm_lines, ret.block_asm_idx, ret.tokens = np.array(new_asm_lines), np.array(new_block_asm_idx), new_tokens
    
    # Datasets will have their cfg's normalized as a list, then will need their tokens recounted
    elif isinstance(cfg_data, (bincfg.CFGDataset, bincfg.MemCFGDataset)):
        using_tokens = None if isinstance(cfg_data, bincfg.CFGDataset) else using_tokens if using_tokens is not None else {}
        ret.cfgs = normalize_cfg_data(cfg_data.cfgs, normalizer=normalizer, inplace=inplace, using_tokens=using_tokens, 
            progress=progress)
        
        if isinstance(cfg_data, bincfg.MemCFGDataset):
            ret.tokens = using_tokens

        # Convert to MemCFGDataset if needed
        if convert_to_mem and isinstance(cfg_data, bincfg.CFGDataset):
            ret = bincfg.MemCFGDataset(ret, normalizer=normalizer, tokens=ret.tokens, keep_memory_addresses=conv_keep_mem_addrs)
    
    else:
        raise TypeError("Got an unknown type: '%s'" % type(cfg_data).__name__)
    
    # Unpack cfg's if needed
    if unpack_cfgs:
        return [ret] if isinstance(ret, (bincfg.CFG, bincfg.MemCFG)) else ret.cfgs

    return ret


def _unpack_cfgs(cfg_data: 'Union[CFGInputDataType, bincfg.CFG, bincfg.MemCFG, bincfg.CFGDataset, bincfg.MemCFGDataset, Iterable]') \
    -> 'tuple[list[Union[bincfg.CFG, bincfg.MemCFG]], list[tuple[int, int]]]':
    """Helper to unpack lists/tuples of cfg's/datasets's, and return their type info

    Will also convert strings to CFG's

    Args:
        cfg_data (Union[str, CFG, MemCFG, CFGDataset, MemCFGDataset, Iterable]): some cfg data. Can be either: str, CFG, MemCFG, 
            CFGDataset, MemCFGDataset, or iterable of previously mentioned types.

    Returns:
        tuple[list[Union[bincfg.CFG, bincfg.MemCFG]], list[tuple[int, int]]]: a 2-tuple of `(cfgs, type_info)`, where 
            `cfgs` is a list of CFG's/MemCFG's containing all of the cfgs that exist in cfg_data in order, and `type_info` 
            is a list of 2-tuples of (start_idx, end_idx) where: `start_idx` is the integer start index of this chunk of 
            data in the `cfgs` list, and `end_idx` is the end index of this chunk of data
    """
    cfgs, cfg_parts = [], []

    # If the user didn't pass an iterable, make it one
    if isinstance(cfg_data, (str, bincfg.CFG, bincfg.MemCFG, bincfg.CFGDataset, bincfg.MemCFGDataset)):
        cfg_data = [cfg_data]
    
    start_idx = 0
    for cfgd in cfg_data:
        
        # CFG's/MemCFG's can simply be added, but Dataset's must have their cfgs lists added
        if isinstance(cfgd, str):
            cfgs.append(bincfg.CFG(cfgd))
        elif isinstance(cfgd, (bincfg.CFG, bincfg.MemCFG)):
            cfgs.append(cfgd)
        elif isinstance(cfgd, (bincfg.CFGDataset, bincfg.MemCFGDataset)):
            cfgs += cfgd.cfgs

        # Update the start_idx and end_idx
        cfg_parts.append((start_idx, len(cfgs)))
        start_idx = len(cfgs)
    
    return cfgs, cfg_parts



"""
Classes for normalizing assembly instructions.
"""

import re
import hashlib
import os
import copy
from types import MethodType
from .norm_utils import imm_to_int, DISINFO_START, DISINFO_END, RE_IMMEDIATE, RE_STRING_LITERAL, parse_disinfo_json, \
    scan_for_token
from ..utils import eq_obj, parameter_saver, paramspec_set_class_funcs, hash_obj
from .base_tokenizer import TokenMismatchError, Tokens, INSTRUCTION_START_TOKEN, TokenizationLevel, \
    parse_tokenization_level, UnknownTokenError
from .base_tokenizer import _USE_DEFAULT_NT


# Default threshold for immediate values for normalization methods such as 'safe'
DEFAULT_IMMEDIATE_THRESHOLD = 5000


# Regexs for capturing immediates and string literals at the start of disassembler info tokens
RE_DISINFO_IMM = re.compile(r'({imm}).*'.format(imm=RE_IMMEDIATE))
RE_DISINFO_STR = re.compile(r'({str}).*'.format(str=RE_STRING_LITERAL))


class MetaNorm(type):
    """A metaclass for BaseNormalizer. 

    The Problem:
        If you change instance functions within the __init__ method (EG: see the SAFE _handle_immediate() function
        being changed in __init__), then 'self' will not automatically be passed to those functions.

        NOTE: this is specifically useful when the effect of a normalization method depends on parameters sent to
        the instance, not inherent to the class

        NOTE: this is not the case for any functions that are set during class initialization (EG: outside of the
        __init__() block)

        So, any functions changed within __init__ methods must be altered to also pass 'self'. I ~could~ force the
        users to have to call a '__post_init__()' function or something, but can we count on them (IE: myself) to
        always do that?...

    The Solution:
        This metaclass inserts extra code before and after any normalizer's __init__ method is called. That code keeps
        track of all instance functions before intitialization, and checks to see if any of them change after
        initialization. This means someone re-set a function within __init__ (IE: self._handle_immediate = ...).
        When this happens, 'self' will not automatically be passed when that function is called. These functions
        are then wrapped to also automatically pass 'self'.

        NOTE: to determine if a function changes, we just check equality between previous and new functions using
        getattr(self, func_name). I don't know why basic '==' works but 'is' and checking id's do not, but I'm not 
        going to question it...

        NOTE: We also have to keep track of the instance functions as an instance variable in case a parent class needs
        their function updated, or if a child class also changes a parent class's function in init

    NOTE: this will mean you cannot call all of that class's methods and expect them to always be the same as calling
    instance methods if you change functions in __init__
    """
    def __new__(cls, name, bases, dct):
        ret_cls = super().__new__(cls, name, bases, dct)  # Create a new class object (not instance)

        old_init = ret_cls.__init__  # Save this class's __init__ function to call later
        def insert_post(self, *args, **kwargs):
            """Create the new __init__ function, inserting code before and after the old __init__"""
            # Keep track of all of this instance's functions. Need to do this as an instance variable in case a parent
            #   class changed things in init so it's not wrapped twice in the child class. Also keep track of which
            #   stack frame needs to remove the __instance_funcs__ attribute
            remove_instance_funcs = False
            if not hasattr(self, '__instance_funcs__') or self.__instance_funcs__ is None:
                self.__instance_funcs__ = {k: getattr(self, k) for k in dir(self) if not k.startswith("__") and callable(getattr(self, k))}
                remove_instance_funcs = True

            # Call the old __init__ function
            parameter_saver(old_init, insert_functions=False)(self, *args, **kwargs)

            # Check if any of the functions before are no longer equal. If so, assume we need to change these functions
            #   to pass self. I don't know why basic '==' works but 'is' and checking id's do not, but I'm not going to
            #   question it...
            new_instance_funcs = {k: getattr(self, k) for k in dir(self) if k in self.__instance_funcs__ and self.__instance_funcs__[k] != getattr(self, k)}
            for k, v in new_instance_funcs.items():
                # Check to make sure v is not already a bound method of self. This can happen if the user sets a method
                #   of self to another previously bound method of self while in __init__
                if isinstance(v, MethodType) and getattr(self, v.__name__) == v:
                    continue

                setattr(self, k, MethodType(v, self))
                self.__instance_funcs__[k] = getattr(self, k)  # Update the instance funcs with the new function
            
            if remove_instance_funcs:
                del self.__instance_funcs__
        
        ret_cls.__init__ = insert_post  # Set this class's __init__ function to be the new one
        return paramspec_set_class_funcs(ret_cls)
    

class NormalizerState:
    """A class that contains information during a normalizer's normalization process"""

    orig_token = None
    """str: The current string token being normalized"""

    token = None
    """str: The current processed version of token if it has already been partially or fully normalized, or None if not"""

    token_type = None
    """str: The token type of the current token, see bincfg.normalization.base_tokenizer.Tokens"""

    token_idx = None
    """int: The index of the current token in 'line'"""

    line = None
    """List[Tuple[str, str, str]]): list of all TokenTuple's in this current line.
       TokenTuple = (token_type [from `bincfg.normalization.base_tokenizer.Tokens` enum], new_token_string, original_token_string)"""
    
    normalized_lines = None
    """List[str]): list of all currently normalized lines/tokens (depending on self.tokenization_level)"""
    
    raw_strings = None
    """List[str]: list of all of the raw strings passed to the current .normalize() call"""
    
    match_instruction_address = None
    """bool: whether or not we are matching instruction addresses at the beginning of assembly lines. This is very likely always True"""
    
    newline_tup = None
    """Optional[Tuple[str, str]]: the newline tuple being used (token_type [probably Tokens.NEWLINE], token_string), or None if not using"""
    
    cfg = None
    """Optional[bincfg.CFG]: the CFG that this token's basic block belongs to, or None if not using"""

    block = None
    """Optional[bincfg.CFGBasicBlock]: the CFGBasicBlock that this token belongs to, or None if not using"""

    memory_start = None
    """Optional[int]: the index of the start of the current memory expression, or None if we are not in a memory expression currently"""

    disinfo_json = None
    """Optional[JSONObject]: the parsed json from a disinfo object"""

    handlers = None
    """Dict[str, Callable[[NormalizerState], Union[str, None]]]: dictionary of current token handler functions"""

    kwargs = None
    """Dict: dictionary of extra kwargs for use in tokenization, or child classes"""

    def __init__(self, **kwargs):
        self.set(**kwargs)
    
    def set(self, **kwargs):
        """Sets the given kwargs on this object's attribute dictionary"""
        for k, v in kwargs.items():
            if k not in dir(self):
                raise ValueError("Cannot set key: %s" % repr(k))
            setattr(self, k, v)
        return self
    
    def copy(self):
        """Returns a copy of this state, but doesn't copy `cfg` or `block`"""
        return NormalizerState(**{k: (copy.deepcopy(v) if k not in ['cfg', 'block'] else v) for k, v in self.__dict__.items()})
    
    def copy_set(self, **kwargs):
        """Copies this state, then updates all the given parameters"""
        ret = self.copy()
        ret.set(**kwargs)
        return ret

    @property
    def token_tuple(self):
        """Returns (token_type, token, orig_token)"""
        return (self.token_type, self.token, self.orig_token)
    
    def __getitem__(self, key):
        """Allows access like dictionary keys"""
        if key in dir(self):
            return getattr(self, key)
        raise KeyError(key)
    
    def __setitem__(self, key, value):
        """Allows access like dictionary keys"""
        if key in dir(self):
            setattr(self, key, value)
        raise KeyError(key)
    
    def __str__(self):
        return repr(self.__dict__)


class BaseNormalizer(metaclass=MetaNorm):
    """A base class for a normalization method. 
    
    This should be subclassed once for each new instruction set to create a base normalizer for that instruction set
    that performs a default 'unnormalized' normalization

    There are three types of functions that are intended to be overridden when needed:

    1. Token handlers: these functions will start with 'handle' and are used to handle either single tokens, or small
       groups of similar tokens (EG: memory expressions). They should accept both self and 'state' as inputs (see
       `bincfg.normalization.base_normalizer.NormalizerState`) and can return either a token which will be added to the
       end of the current line, or None to not add any token post-calling.
    2. Opcode handlers: these functions will start with 'opcode' and are used to handle specific opcodes (not the
       'opcode' token in general, only specific ones like 'call' or 'jump' opcodes). They should accept both self and 
       'state' as inputs (See ``bincfg.normalization.base_normalizer.NormalizerState``) and can return either the integer 
       index of the next token that should be checked (IE: "we have handled all tokens up to but not including this index"),
       or None to indicate the previously mentioned index is just one after the opcode. These operate directly on the
       state's current '.line' attribute. These are expected to be called only after the entire current line has finished
       being parsed and normalized. New opcode handlers can be added with self.register_opcode_handler()
    3. Administrative functions: these functions perform different administrative operations before, during, or after
       normalizing the individual tokens. Some examples include:

       - 'finalize_instruction': used as a post-processing function once an instruction has finished being normalized to perform
         extra processing to the line, apply opcode handlers, stringify the line, update the normalizer state
       - 'hash_token': hashes a fully processed string token (if self.anonymize_tokens=True)
       - 'stringify_line': takes the current line of token tuples and converts into strings based on self.tokenization_level
    
    Disassembler Information:

    Extra information from the disassembler can be inserted into the lines within angle brackets "<>" (see
    :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer` for info on how this can be tokenized). This disassembler
    info will be treated as a single token, and passed to the `self.handle_disassembler_info` function. By default, the
    normalizer will check for the following in order

        1. Valid JSON. If the data inside of the angle brackets is valid JSON, then it will be parsed into a JSON object.
           This JSON object will be inserted into the `state.disinfo_json` attribute in the normalizer state. There are
           a few special cases for this JSON data that have special effects by default:

            * If this object is an integer, we will attempt to insert it into a previous immediate value like in #2 below
            * If this is a string, we will always insert it as a string literal like in #3 below
            * If this is a dictionary, there are a few special keys that one can use:

                - 'immediate': value should be an integer. We will attempt to insert value into a previous immediate
                  value like in #2 below
                - 'insert': this value will be inserted into the string. If it is already a string, it is left as-is. If
                  not a string, then we call repr() on it to convert it into a string. Insertion actions depend on whether 
                  or not the key 'insert_type' is present. 
                  
                  If not present, this value will first be tokenized/normalized by this normalizer and that
                  value + token type will be inserted. Should that fail, then the value will be inserted as a string
                  literal WITHOUT processing it as a string literal token (and, it won't have quotes on it). 
                  
                  If the 'insert_type' key is present, then it can be one of two values:

                    * String token_type: the value will be handled as if it is of this token type, no matter what the
                      value actually is, then it will be inserted (assuming that token handler did not return None)
                    * False (the JSON object, not the string): the value will be immediately inserted as a string literal
                      WITHOUT processing it as a string literal token (and, it won't have quotes on it)
                
                - 'insert_type': Determines the token type for an 'insert' key value. Ignored if the 'insert' key is not
                  present. See the 'insert' key for more info
            
        2. Otherwise, if the disassembler info token starts with an immediate value within the angle brackets, and there
           is an immediate value token immediately preceeding them (ignoring spacing tokens), this will replace said 
           immediate value token with the immediate value found within the disassembler info. The inserted value will
           first be handled by the appropriate handler for Token.IMMEDIATE token types.
           EG: "add rax 0xffff <-1>"  ->  "add rax -1"
        3. Otherwise, if the disassembler info token starts with a string literal, this will insert that string literal
           right where it appears (and, that string literal will be handled with `self.handle_string_literal`). The inserted
           value will first be handled by the appropriate handler for Token.STRING_LITERAL token types.
        4. Finally, if it doesn't match anything above, then it will fail silently and be ignored. If you wish to raise
           an error when this happens instead, you can pass `raise_unk_di=True` when calling `.normalize()`
    
    The disassembler tokens themselves are always ignored by default.

    NOTE: escapes will be treated normally within all strings. EG: '\\n' will be considered the newline character, but
    '\\\\n' will escape the escape and produce the string '\\n'. 
    
    NOTE: immediates and string literals must match those found in ``bincfg.normalization.norm_utils`` (`RE_IMMEDIATE`
    and `RE_STRING_LITERAL`). The disassembler info does not take into account the regex's used to parse immediates
    and string literals for the specific normalizer.
       
    
    Parameters
    ----------
    tokenizer: `Tokenizer`
        the tokenizer to use
    token_handlers: `Optional[Dict[str, Callable[[NormalizerState], Union[None, str]]]]`
        optional dictionary mapping string token types to functions to handle those tokens. These will override any
        token handlers that are used by default (IE: all of the `self.handle_*` functions). Functions should take one
        arg (the current normalizer state) as input and return either the next string token to add to the current line,
        or None to not add anything. This is useful for adding more methods to handle new token types that are not builtin.
    token_sep: `Optional[str]`
        the string to use to separate each token in returned instruction lines. Only used if tokenization_level is 
        'instruction'. If None, then a default value will be used (' ' for unnormalized using BaseNormalizer(), '_' 
        for everything else)
    tokenization_level: `Optional[Union[TokenizationLevel, str]]`
        the tokenization level to use for return values. Can be a string, or a ``TokenizationLevel`` type. Strings can be:

            - 'op': tokenized at the opcode/operand level. Will insert a 'INSTRUCTION_START' token at the beginning of
              each instruction line
            - 'inst'/'instruction': tokenized at the instruction level. All tokens in each instruction line are joined
              together using token_sep to construct the final token
            - 'auto': pick the default value for this normalization technique

    anonymize_tokens: `bool`
        if True, then tokens will be annonymized by taking their 4-byte shake_128 hash. Why does this exist? Bureaucracy.
    """

    DEFAULT_TOKENIZATION_LEVEL = TokenizationLevel.INSTRUCTION
    """The default tokenization level used for this normalizer"""

    renormalizable = False
    """Whether or not this normalization method can be renormalized later by other normalization methods"""

    tokenizer = None
    """The tokenizer used for this normalizer"""

    token_sep = None
    """The separator string used for this normalizer
    
    Will default to ' '
    """

    tokenization_level = TokenizationLevel.AUTO
    """The tokenization level to use for this normalizer"""

    def __init__(self, tokenizer, token_handlers=None, token_sep=' ', tokenization_level=TokenizationLevel.AUTO, anonymize_tokens=False):
        self.tokenizer, self.token_sep, self.anonymize_tokens = tokenizer, token_sep, anonymize_tokens
        self.token_handlers = {} if token_handlers is None else token_handlers
        self.tokenization_level = parse_tokenization_level(tokenization_level, self.DEFAULT_TOKENIZATION_LEVEL)

        self.opcode_handlers = []
    
    def register_opcode_handler(self, op_regex, func_or_str_name):
        """Registers an opcode handler for this normalizer

        Adds the given `op_regex` as an opcode to handle during self._handle_instruction() along with the given function
        to call with token/cfg arguments. `op_regex` can be either a compiled regex expression, or a string which
        will be compiled into a regex expression. `func_or_str_name` can either be a callable, or a string. If it's
        a string, then that attribute will be looked up on this normalizer dynamically to find the function to use.

        Notes for registering opcode handlers:

            1. passing instance method functions converts them to strings automatically
            2. passing lambda's or inner functions (not at global scope) would not be able to be pickled
            3. opcodes will be matched in the order they were passed in

        Args:
            op_regex (Union[str, Pattern]): a string or compiled regex
            func_or_str_name (Union[Callable, str]): the function to call with token/cfg arguments when an opcode 
                matches op_regex, or a string name of a callable attribute of this normalizer to be looked up dynamically
        """
        op_regex = re.compile(op_regex) if isinstance(op_regex, str) else op_regex

        if not isinstance(func_or_str_name, str):
            # Check it is callable
            if not callable(func_or_str_name):
                raise TypeError("fun_or_str_name must be str or callable, not '%s'" % type(func_or_str_name))

            # Check if the passed function is an instance method of this normalization method class specifically
            # Have to check if func_or_str_name has a __name__ attribute first since they could sometimes be _LOF classes
            if hasattr(func_or_str_name, '__name__') and hasattr(self.__class__, func_or_str_name.__name__):
                func_or_str_name = func_or_str_name.__name__

        self.opcode_handlers.append((op_regex, func_or_str_name))
    
    def tokenize(self, *strings, newline_tup=_USE_DEFAULT_NT, match_instruction_address=True, **kwargs):
        """Tokenizes the given strings using this normalizer's tokenizer

        See the docs for :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer` for more info on how tokenization
        works, how to create subclasses, etc.

        Args:
            strings (str): arbitrary number of strings to tokenize.
            newline_tup (Optional[Tuple[str, str]]): the tuple to insert inbetween each passed string, or None to not 
                insert anything. Defaults to `self.__class__.DEFAULT_NEWLINE_TUPLE`.
            match_instruction_address (bool, optional): if True, will match instruction addresses. If there is an immediate
                value at the start of a line (IE: start of a string in `strings`, or immediately after a Tokens.NEWLINE
                or Tokens.INSTRUCTION_START [ignoring any Tokens.SPACING]), then that token will be converted into a
                Tokens.INSTRUCTION_ADDRESS token. If there is a Tokens.COLON immediately after that token (again, ignoring
                any Tokens.SPACING), then that first Tokens.COLON match will be appended (along with any inbetween Tokens.SPACING)
                to that Tokens.INSTRUCTION_ADDRESS token. For example, using the x86 tokenization scheme:

                    - "0x1234: add rax rax" -> [(Tokens.INSTRUCTION_ADDRESS, '0x1234:'), ...]
                    - "  0x1234     : add rax rax" -> [(Tokens.SPACING, '  '), (Tokens.INSTRUCTION_ADDRESS, '0x1234     :'), ...]
                    - "0x1234 add rax rax" -> [(Tokens.INSTRUCTION_ADDRESS, '0x1234'), ...]
                
            kwargs (Any): extra kwargs to store in the tokenizer state, for use in child classes

        Returns:
            List[Tuple[str, str]]: list of (token_type, token) tuples
        """
        return self.tokenizer(*strings, newline_tup=newline_tup, match_instruction_address=match_instruction_address, **kwargs)

    def normalize(self, *strings, cfg=None, block=None, newline_tup=_USE_DEFAULT_NT, match_instruction_address=True, **kwargs):
        """Normalizes the given iterable of strings.

        Args:
            strings (str): arbitrary number of strings to normalize
            cfg (Union[CFG, MemCFG], optional): either a ``CFG`` or ``MemCFG`` object that these lines occur 
                in. Used for determining function calls to self, internal functions, and external functions. If not 
                passed, then these will not be used. Defaults to None.
            block (Union[CFGBasicBlock, int], optional): either a ``CFGBasicBlock`` or integer block_idx in a ``MemCFG``
                object. Used for determining function calls to self, internal functions, and external functions. If not 
                passed, then these will not be used. Defaults to None.
            newline_tup (Tuple[str, str], optional): the tuple to insert inbetween each passed string, or None to not 
                insert anything. Defaults to self.tokenizer.DEFAULT_NEWLINE_TUPLE
            match_instruction_address (bool, optional): if True, will match instruction addresses. If there is an immediate
                value at the start of a line (IE: start of a string in `strings`, or immediately after a Tokens.NEWLINE
                or Tokens.INSTRUCTION_START [ignoring any Tokens.SPACING]), then that token will be converted into a
                Tokens.INSTRUCTION_ADDRESS token. If there is a Tokens.COLON immediately after that token (again, ignoring
                any Tokens.SPACING), then that first Tokens.COLON match will be appended (along with any inbetween Tokens.SPACING)
                to that Tokens.INSTRUCTION_ADDRESS token. For example, using the x86 tokenization scheme:

                    - "0x1234: add rax rax" -> [(Tokens.INSTRUCTION_ADDRESS, '0x1234:'), ...]
                    - "  0x1234     : add rax rax" -> [(Tokens.SPACING, '  '), (Tokens.INSTRUCTION_ADDRESS, '0x1234     :'), ...]
                    - "0x1234 add rax rax" -> [(Tokens.INSTRUCTION_ADDRESS, '0x1234'), ...]
                
            kwargs (Any): extra kwargs to pass along to tokenization method, and to store in normalizer state

        Returns:
            List[str]: a list of normalized string instruction lines
        """
        # If no strings were passed, return empty list
        if len(strings) == 0:
            return []
        
        # Check if the first string is an instruction start. If so, we are normalizing an already-normalized string that
        #   was normalized with tokenization_level='op'. Combine all the strings together, assuming the instruction start
        #   tokens are inbetween all of the instructions
        if strings[0] == INSTRUCTION_START_TOKEN:
            newline_tup = None
        
        newline_tup = newline_tup if newline_tup is not _USE_DEFAULT_NT else self.tokenizer.DEFAULT_NEWLINE_TUPLE
        
        # Get the current mapping of token types to their handler functions
        handler_mapping = {

            Tokens.INSTRUCTION_ADDRESS: self.handle_instruction_address,
            #Tokens.INSTRUCTION_START: Will occurr, but we handle it seperately,
            #Tokens.SPLIT_IMMEDIATE: Should never occurr,
            Tokens.DISASSEMBLER_INFO: self.handle_disassembler_info,
            Tokens.NEWLINE: self.handle_newline,
            Tokens.SPACING: self.handle_spacing,

            Tokens.OPEN_BRACKET: self.handle_all_symbols,
            Tokens.CLOSE_BRACKET: self.handle_all_symbols,
            Tokens.PLUS_SIGN: self.handle_all_symbols,
            Tokens.TIMES_SIGN: self.handle_all_symbols,
            Tokens.COLON: self.handle_all_symbols,

            Tokens.INSTRUCTION_PREFIX: self.handle_instruction_prefix,
            Tokens.OPCODE: self.handle_opcode,
            Tokens.REGISTER: self.handle_register,
            Tokens.IMMEDIATE: self.handle_immediate,

            Tokens.MEMORY_SIZE: self.handle_memory_size,
            #Tokens.MEMORY_EXPRESSION: doesn't occurr here, that's for subclasses
            Tokens.BRANCH_PREDICTION: self.handle_branch_prediction,
            Tokens.STRING_LITERAL: self.handle_string_literal,

            Tokens.MISMATCH: self.handle_mismatch,
        }
        handler_mapping.update(self.token_handlers)

        # Initialize the current state that gets passed around to function calls
        state = NormalizerState(cfg=cfg, block=block, newline_tup=newline_tup, match_instruction_address=match_instruction_address,
                                normalized_lines=[], line=[], raw_strings=strings, kwargs=kwargs, handlers=handler_mapping)
        
        for token_type, old_token in self.tokenize(*strings, newline_tup=newline_tup, match_instruction_address=match_instruction_address, **kwargs):
            state.token_type, state.orig_token, state.token = token_type, old_token, old_token
            
            # Handle this current token
            self._handle_token(state)
            
            # If this was a newline token or instruction start token, call our line handler
            if state.token_type in [Tokens.NEWLINE, Tokens.INSTRUCTION_START]:
                self.finalize_instruction(state)
                self.add_line_to_sentence(state)
        
        self.finalize_instruction(state)
        self.add_line_to_sentence(state)

        # If we currently have no lines, then insert an empty string
        if len(state.normalized_lines) == 0:
            state.normalized_lines.append("")

        # If we are anonymizing the tokens, do that now
        if self.anonymize_tokens:
            for i, t in enumerate(state.normalized_lines):
                state.normalized_lines[i] = self.hash_token(t)
        
        return state.normalized_lines
    
    def _handle_token(self, state, insert_at=None, insert_token=True):
        """Handles a single token of the given token_type. Returns the state"""
        state.token = state.handlers[state.token_type](state) if state.token_type in state.handlers \
            else None if state.token_type in [Tokens.INSTRUCTION_START] \
            else self.handle_unknown_token(state)

        # If we are inserting the token
        if insert_token:
            # If we have a location to insert at
            if insert_at is not None:
                if state.token is not None:
                    state.line[insert_at] = (state.token_type, state.token, state.orig_token)
                else:
                    state.line = state.line[:insert_at] + state.line[insert_at + 1:]
                
            # Otherwise just append
            elif state.token is not None:
                state.line.append((state.token_type, state.token, state.orig_token))

        return state
    
    def handle_opcode(self, state):
        """Handles an opcode. Defaults to returning the original token

        NOTE: This should only be used to determine how all opcode strings are handled. For how to handle specific opcodes
        to give them different behaviors, see :func:`~bincfg.normalization.base_normalizer.BaseNormalizer.register_opcode_handler`

        Should return either the token to add to the current line, or None to not add any token

        Args:
            state (NormalizerState): dictionary of current state information. See ``bincfg.normalization.base_normalizer.NormalizerState``
        """
        return state.token
    
    def handle_all_symbols(self, state):
        """Handles symbols ('+', '[', ']', '*', ':'). Defaults to returning the original token

        Should return either the token to add to the current line, or None to not add any token

        Args:
            state (NormalizerState): dictionary of current state information. See ``bincfg.normalization.base_normalizer.NormalizerState``
        """
        return state.token
    
    def handle_memory_size(self, state):
        """Handles a memory size. Defaults to returning the original token

        Should return either the token to add to the current line, or None to not add any token

        Args:
            state (NormalizerState): dictionary of current state information. See ``bincfg.normalization.base_normalizer.NormalizerState``
        """
        return state.token
    
    def handle_register(self, state):
        """Handles a register. Defaults to returning the original token

        Should return either the token to add to the current line, or None to not add any token

        Args:
            state (NormalizerState): dictionary of current state information. See ``bincfg.normalization.base_normalizer.NormalizerState``
        """
        return state.token
    
    def handle_instruction_prefix(self, state):
        """Handles an instruction prefix. Defaults to returning the original token

        Should return either the token to add to the current line, or None to not add any token

        Args:
            state (NormalizerState): dictionary of current state information. See ``bincfg.normalization.base_normalizer.NormalizerState``
        """
        return state.token
    
    def handle_branch_prediction(self, state):
        """Handles a branch prediction. Defaults to returning the original token

        Should return either the token to add to the current line, or None to not add any token

        Args:
            state (NormalizerState): dictionary of current state information. See ``bincfg.normalization.base_normalizer.NormalizerState``
        """
        return state.token
    
    def handle_instruction_address(self, state):
        """Handles an instruction address. Defaults to ignoring these tokens

        Should return either the token to add to the current line, or None to not add any token

        Args:
            state (NormalizerState): dictionary of current state information. See ``bincfg.normalization.base_normalizer.NormalizerState``
        """
        return None
    
    def handle_spacing(self, state):
        """Handles spacing. Defaults to ignoring these tokens

        Should return either the token to add to the current line, or None to not add any token

        Args:
            state (NormalizerState): dictionary of current state information. See ``bincfg.normalization.base_normalizer.NormalizerState``
        """
        return None

    def handle_immediate(self, state):
        """Handles an immediate value. Defaults to converting into decimal

        Should return either the token to add to the current line, or None to not add any token

        Args:
            state (NormalizerState): dictionary of current state information. See ``bincfg.normalization.base_normalizer.NormalizerState``
        """
        return str(imm_to_int(state.token))
    
    def handle_newline(self, state):
        """Handles a newline token. Defaults to ignoring the token

        Args:
            state (NormalizerState): dictionary of current state information. See ``bincfg.normalization.base_normalizer.NormalizerState``
        """
        return None
    
    def handle_disassembler_info(self, state):
        """Handles disassembler information

        See :func:`~bincfg.normalization.base_normalizer.BaseNormalizer` for more info on how disassembler info is parsed.

        Should return either the token to add to the current line, or None to not add any token

        Args:
            state (NormalizerState): dictionary of current state information. See ``bincfg.normalization.base_normalizer.NormalizerState``
        """
        # Clear the disinfo_json attribute
        state.disinfo_json = None

        # Get the actual token
        disinfo = state.token[len(DISINFO_START):-len(DISINFO_END)]

        # Functions for inserting immediate values and string literals
        def _insert_imm(val):
            idx = scan_for_token(state.line, type=[Tokens.IMMEDIATE], stop_unmatched=True, ignore_type=[Tokens.SPACING], start=-1, increment=-1)
            if idx is not None:
                state.line = self._handle_token(state.copy_set(token_type=Tokens.IMMEDIATE, token=val, orig_token=val), insert_at=idx).line

        def _insert_str(val):
            state.line = self._handle_token(state.copy_set(token_type=Tokens.STRING_LITERAL, token=val, orig_token=val)).line

        # Attempt to parse as a JSON object
        parsed_json = parse_disinfo_json(disinfo)
        if parsed_json is not None:
            state.disinfo_json = parsed_json

            # If this is an immediate or a string, apply those
            if isinstance(parsed_json, int):
                _insert_imm(str(parsed_json))
            elif isinstance(parsed_json, str):
                _insert_str(_norm_str('"' + parsed_json + '"'))
            
            # If this is a dictionary with special keys, handle those
            elif isinstance(parsed_json, dict):
                if 'immediate' in parsed_json:
                    _insert_imm(str(parsed_json['immediate']))
                elif 'insert' in parsed_json:
                    json_insert = parsed_json['insert']

                    if 'insert_type' in parsed_json:
                        if isinstance(parsed_json['insert_type'], bool) and not parsed_json['insert_type']:
                            # Leave as-is if a string, otherwise convert to string with repr()
                            ins_str = json_insert if isinstance(json_insert, str) else repr(json_insert)
                            state.line.append((Tokens.STRING_LITERAL, json_insert, json_insert))
                        else:
                            # Insert string quotes if this is a string literal type, otherwise call repr() to convert to string
                            ins_str = _norm_str('"' + json_insert + '"') if parsed_json['insert_type'] in [Tokens.STRING_LITERAL] else repr(json_insert)
                            self._handle_token(state.set(token=ins_str, token_type=parsed_json['insert_type'], orig_token=ins_str))
                    else:
                        # Leave as-is if a string, otherwise convert to string with repr()
                        ins_str = json_insert if isinstance(json_insert, str) else repr(json_insert)
                        tokens = self.tokenize(ins_str, newline_tup=None, match_instruction_address=False, **state['kwargs'])
                        if len(tokens) != 1:
                            raise ValueError("Could not tokenize insert value: %s" % ins_str)
                        self._handle_token(state.set(token=tokens[0][1], token_type=tokens[0][0], orig_token=ins_str))

        else:
            # Check for an immediate value at the start
            mo_imm = RE_DISINFO_IMM.fullmatch(disinfo)
            mo_str = RE_DISINFO_STR.fullmatch(disinfo)
            if mo_imm is not None:
                _insert_imm(mo_imm.groups()[0])

            # Check for a string literal
            elif mo_str is not None:
                _insert_str(_norm_str(mo_str.groups()[0]))
            
            # Finally, check if we should raise an error due to unknown disassembler info
            elif "raise_unk_di" in state.kwargs and state.kwargs['raise_unk_di']:
                raise ValueError("Unknown disassembler info: %s" % repr(disinfo))
            
        return None
    
    def handle_string_literal(self, state):
        """Handles string literals. Defaults to returning the original token as a double-quoted string

        Should return either the token to add to the current line, or None to not add any token

        Args:
            state (NormalizerState): dictionary of current state information. See ``bincfg.normalization.base_normalizer.NormalizerState``
        """
        return _norm_str(state.token)

    def handle_mismatch(self, state):
        """What to do when the normalizaion method finds a token mismatch (in case they were ignored in the tokenizer)

        Defaults to raising a TokenMismatchError()

        Should return either the token to add to the current line, or None to not add any token

        Args:
            state (NormalizerState): dictionary of current state information. See ``bincfg.normalization.base_normalizer.NormalizerState``
        
        Raises:
            TokenMismatchError: always
        """
        raise TokenMismatchError("Mismatched token %s found during normalization!" % repr(state.token))
    
    def handle_unknown_token(self, state):
        """Handles an unknown token. Defaults to raising an UnknownTokenError

        Should return either the token to add to the current line, or None to not add any token

        Args:
            state (NormalizerState): dictionary of current state information. See ``bincfg.normalization.base_normalizer.NormalizerState``
        
        Raises:
            UnknownTokenError: always
        """
        raise UnknownTokenError("Unknown token type %s" % repr(state.token_type))
    
    def finalize_instruction(self, state):
        """Handles an entire instruction once reaching a new line

        If overridden, should at the very least:

            - call all the registered opcode handlers for each known opcode token (while updating token_type/token/token_idx)
        
        By default, each opcode handler is expected to take in the current state, and return either the integer index
        of the next token that should be checked (IE: "we have handled all tokens up to but not including this index"),
        or None to indicate the previously mentioned index is just one after the opcode

        Args:
            state (NormalizerState): dictionary of current state information. See ``bincfg.normalization.base_normalizer.NormalizerState``
        """
        # If this is an empty line, just return
        if len(state.line) == 0:
            return
        
        # Handle all of the opcodes
        idx = 0
        while idx < len(state.line):
            
            # Check for any handled opcodes
            if state.line[idx][0] == Tokens.OPCODE:
                state.token_type, state.token, _ = state.line[idx]
                state.token_idx = idx

                for regex, func in self.opcode_handlers:
                    if regex.fullmatch(state.token) is not None:
                        # Check for string name to lookup on self
                        if isinstance(func, str):
                            func = getattr(self, func)

                        new_idx = func(state)
                        idx = (new_idx - 1) if new_idx is not None else idx
                        break
            
            idx += 1
    
    def add_line_to_sentence(self, state):
        """Stringifies the current line, then adds it to the normalized lines and clears state.line"""
        sl = self.stringify_line(state)
        state.normalized_lines += [sl] if isinstance(sl, str) else list(sl)
        state.line.clear()
    
    def hash_token(self, token):
        """Hashes tokens during annonymization

        By default, converts each individual token into its 4-byte shake_128 hash

        Args:
            token (str): the string token to hash
        
        Returns:
            str: the 4-byte shake_128 hash of the given token
        """
        hasher = hashlib.shake_128()
        hasher.update(token.encode('utf-8'))
        return hasher.hexdigest(4)

    def stringify_line(self, state):
        """Converts the current line into a list of final normalized string tokens and returns that list

        Also normalizes the case, converting all tokens (except those in strings) to lowercase

        Args:
            state (NormalizerState): dictionary of current state information. See ``bincfg.normalization.base_normalizer.NormalizerState``
            
        Returns:
            List[str]: a list of tokens to add to state.normalized_lines
        """
        tokens = [_norm_case(t, n) for n, t, _ in state.line]
        if len(tokens) == 0:
            return []
        
        if self.tokenization_level == TokenizationLevel.INSTRUCTION:
            return [self.token_sep.join(tokens)]
        elif self.tokenization_level == TokenizationLevel.OPCODE:
            return [INSTRUCTION_START_TOKEN] + tokens
        else:
            raise ValueError("Unknown TokenizationLevel: %s" % self.tokenization_level)
    
    def __call__(self, *strings, cfg=None, block=None, newline_tup=_USE_DEFAULT_NT, match_instruction_address=True, **kwargs):
        """Normalizes the given iterable of strings.

        Args:
            strings (str): arbitrary number of strings to normalize
            cfg (Union[CFG, MemCFG], optional): either a ``CFG`` or ``MemCFG`` object that these lines occur 
                in. Used for determining function calls to self, internal functions, and external functions. If not 
                passed, then these will not be used. Defaults to None.
            block (Union[CFGBasicBlock, int], optional): either a ``CFGBasicBlock`` or integer block_idx in a ``MemCFG``
                object. Used for determining function calls to self, internal functions, and external functions. If not 
                passed, then these will not be used. Defaults to None.
            newline_tup (Tuple[str, str], optional): the tuple to insert inbetween each passed string, or None to not 
                insert anything. Defaults to self.tokenizer.DEFAULT_NEWLINE_TUPLE
            match_instruction_address (bool, optional): if True, will match instruction addresses. If there is an immediate
                value at the start of a line (IE: start of a string in `strings`, or immediately after a Tokens.NEWLINE
                or Tokens.INSTRUCTION_START [ignoring any Tokens.SPACING]), then that token will be converted into a
                Tokens.INSTRUCTION_ADDRESS token. If there is a Tokens.COLON immediately after that token (again, ignoring
                any Tokens.SPACING), then that first Tokens.COLON match will be appended (along with any inbetween Tokens.SPACING)
                to that Tokens.INSTRUCTION_ADDRESS token. For example, using the x86 tokenization scheme:

                    - "0x1234: add rax rax" -> [(Tokens.INSTRUCTION_ADDRESS, '0x1234:'), ...]
                    - "  0x1234     : add rax rax" -> [(Tokens.SPACING, '  '), (Tokens.INSTRUCTION_ADDRESS, '0x1234     :'), ...]
                    - "0x1234 add rax rax" -> [(Tokens.INSTRUCTION_ADDRESS, '0x1234'), ...]
                
            kwargs (Any): extra kwargs to pass along to tokenization method, and to store in normalizer state

        Returns:
            List[str]: a list of normalized string instruction lines
        """
        return self.normalize(*strings, cfg=cfg, block=block, newline_tup=newline_tup, match_instruction_address=match_instruction_address, **kwargs)

    def __eq__(self, other):
        """Checks equality between this normalizer and another. 
        
        Defaults to checking if class types, tokenizers, and tokenization_level are the same. Future children should 
            also check any kwargs.
        """
        return type(self) == type(other) and all(eq_obj(self, other, selector=s) for s in 
            ['tokenizer', 'tokenization_level', 'anonymize_tokens', 'renormalizable', 'token_sep', 'token_handlers', 'opcode_handlers'])
    
    def __hash__(self):
        return hash_obj([type(self).__name__, [r for r, _ in self.opcode_handlers], self.tokenizer, self.tokenization_level.name,
                         self.anonymize_tokens, self.renormalizable, self.token_sep, self.token_handlers], return_int=True)
    
    def __repr__(self) -> str:
        _num_str_chars = 30
        def _clean_str(o):
            s = repr(o)
            if len(s) > _num_str_chars:
                return o.__class__.__name__ + "(...)"
            return s
        def _clean_kwarg(k, v):
            if k == 'tokenizer':
                ret = self.tokenizer
            elif k == 'token_sep':
                ret = self.token_sep
            elif k == 'tokenization_level':
                ret = self.tokenization_level.name.lower()
            else:
                ret = v
            return _clean_str(ret)

        args_kwargs_str = [('%s=%s' % (k, _clean_str(v))) for k, v in self.__savedparams__['__init__']['args'].items()] + \
            [('%s=%s' % (k, _clean_kwarg(k, v))) for k, v in self.__savedparams__['__init__']['kwargs'].items()]
        return self.__class__.__name__ + "(" + ', '.join(args_kwargs_str) + ")"
    
    def __str__(self) -> str:
        return self.__class__.__name__.lower() + (('_op' if self.tokenization_level == TokenizationLevel.OPCODE else '_inst') if self.tokenization_level != self.DEFAULT_TOKENIZATION_LEVEL else '')


def _norm_case(token, token_type):
    """Converts token to lowercase, unless it is a string token_type"""
    return token.lower() if token_type not in [Tokens.STRING_LITERAL] else token


def _norm_str(token):
    """Normalizes a string token

    `token` should be the string token, including the starting/ending quotes (can be either single or double quotes,
    but must be matching).

    Strings are normalized for readability. This can handle unicode, escaped characters, etc. Strings are treated
    how python would treat them (regarding how to escape characters, etc.).

    The general workflow is:

        1. Encode the string into 'utf-8' bytes, then convert to a string. This forces python to convert any weird
           characters (tabs, newlines, unicode, extra quotes, ...) into escaped format.
        2. Remove the starting quote + 'b' and ending quote to get the original string in this new format
        3. Un-escape any doubly escaped '\\' characters that were escaped when converting into 'utf-8'
        4. Replace any escaped single quotes with plain single quotes, but only under certain conditions, see the note below.
        5. Remove the old starting/ending quotes
        6. Go through the string finding any un-escaped double quotes and escape them
        7. Replace any escaped single quotes with plain single quotes since we only use double quotes as outer quotes
    
    NOTE: As far as I can tell, calling str() on the encoded string functions much like calling repr() on a string. It will
    default to wrapping the string in single quotes, UNLESS that string contains single quotes in which case it will
    wrap in double quotes so as to not have to escape the single quotes, UNLESS-UNLESS the string also contains
    double quotes in which case it will wrap in single quotes and escape all inner single quotes. So, if the string
    contains both single and double quotes, we'll have to un-escape any single quotes within the string to keep
    everything normalized as expected.

    This should be able to normalize any weird combinations of characters/spacing/unicode/etc., and force strings to
    start/end with double quotes.
    """
    encoded = str(token.encode('utf-8'))[2:-1].replace('\\\\', '\\')
    
    # Remove extra escapes possibly added
    if '"' in encoded and "'" in encoded:
        encoded = encoded.replace("\\'", "'")
    
    # Check escaped characters to find un-escaped double quotes, and escape them
    escaped = ""
    last_escape = False
    for c in encoded[1:-1]:
        # If this is a double quote that was not escaped, add in an escape character
        if c == '"' and not last_escape:
            escaped += '\\'
        
        # If the last character was an escape character, turn off last_escape. Otherwise if this 
        #   is an escape character, turn on last_escape
        last_escape = False if last_escape else (c == '\\')

        # Add in this character always
        escaped += c

    return '"' + escaped.replace("\\'", "'") + '"'


# Libc function names gathered from: https://www.gnu.org/software/libc/manual/html_node/Function-Index.html
# Code used to generate these from raw copy/pasted website data:
"""
import re

libc_funcs = set()
for l in s.split('\n'):
    mo = re.fullmatch(r'[ \t\n]*([0-9a-zA-Z_*]+):.*', l)

    if mo is not None:
        libc_funcs.add(mo.groups()[0])

with open('./libc_func_names.txt', 'w') as f:
    for n in sorted(list(libc_funcs)):
        f.write(n + '\n')
"""
from .libc_func_names import FUNC_NAMES
LIBC_FUNCTION_NAMES = set([n.replace('\n', '') for n in FUNC_NAMES.split('\n') if not re.fullmatch(r'[ \t\n]*', n)])



"""
An assortment of helper/utility functions for tokenization/normalization.
"""

import bincfg
import re
import traceback
import json
from functools import reduce


# The string/char that is inserted at the start and end of special tokens to signify them as special (and to make sure
#   they don't conflict with other actual tokens and whatnot)
SPECIAL_TOKEN_DESIGNATOR = '#'

# Token to insert at the start of each instruction for opcode-level tokenization
INSTRUCTION_START_TOKEN = '{0}start_instr{0}'.format(SPECIAL_TOKEN_DESIGNATOR)

# Token that is used to identify the start of a split immediate token
SPLIT_IMMEDIATE_TOKEN = '{0}split_imm{0}'.format(SPECIAL_TOKEN_DESIGNATOR)

# Constant string values for normalized tokens
STRING_LITERAL_STR = '{0}str{0}'.format(SPECIAL_TOKEN_DESIGNATOR)
IMMEDIATE_VALUE_STR = '{0}immval{0}'.format(SPECIAL_TOKEN_DESIGNATOR)
FUNCTION_CALL_STR = '{0}func{0}'.format(SPECIAL_TOKEN_DESIGNATOR)
RECURSIVE_FUNCTION_CALL_STR = '{0}self{0}'.format(SPECIAL_TOKEN_DESIGNATOR)
INTERNAL_FUNCTION_CALL_STR = '{0}innerfunc{0}'.format(SPECIAL_TOKEN_DESIGNATOR)
EXTERNAL_FUNCTION_CALL_STR = '{0}externfunc{0}'.format(SPECIAL_TOKEN_DESIGNATOR)
MULTI_FUNCTION_CALL_STR = '{0}multifunc{0}'.format(SPECIAL_TOKEN_DESIGNATOR)
JUMP_DESTINATION_STR = '{0}jmpdst{0}'.format(SPECIAL_TOKEN_DESIGNATOR)
MEMORY_EXPRESSION_STR = '{0}memexpr{0}'.format(SPECIAL_TOKEN_DESIGNATOR)
GENERAL_REGISTER_STR = '{0}reg{0}'.format(SPECIAL_TOKEN_DESIGNATOR)
SEGMENT_STR = '{0}seg{0}'.format(SPECIAL_TOKEN_DESIGNATOR)
SEGMENT_ADDRESS_STR = '{0}segaddr{0}'.format(SPECIAL_TOKEN_DESIGNATOR)
MEM_SIZE_TOKEN_STR = '{0}memptr{0}'.format(SPECIAL_TOKEN_DESIGNATOR)
DISPLACEMENT_IMMEDIATE_STR = '{0}dispmem{0}'.format(SPECIAL_TOKEN_DESIGNATOR)
LARGE_BLOCK_STR = '{0}large_block{0}'.format(SPECIAL_TOKEN_DESIGNATOR)


# Some helpful regular expressions

# Possible immediate values: hexadecimal, octal, decimal (NOTE: the need to be processed in that order so decimal immediate
#    doesn't take up the initial '0' in front of hexadecimal/octal)
RE_IMM_HEX = r'-?0x[0-9a-f]+'
RE_IMM_OCT = r'-?0o[0-7]+'
RE_IMM_BIN = r'-?0b[01]+'
RE_IMM_INT = r'-?[0-9]+'
RE_IMMEDIATE = r'(?:{hex}|{oct}|{bin}|{int})'.format(hex=RE_IMM_HEX, oct=RE_IMM_OCT, int=RE_IMM_INT, bin=RE_IMM_BIN)

# Match string literals. Allows for strings starting/ending with either single or double quotes, and one can escape
#   quotes with \' or \", and can escape the escape with \\
# Also matches '#str#' as a string literal
RE_STRING_LITERAL = r'(?:"[^"\\]*(?:\\.[^"\\]*)*"|\'[^\'\\]*(?:\\.[^\'\\]*)*\')|{str}'.format(str=STRING_LITERAL_STR)

# Various symbol characters
RE_PLUS_SIGN = r'\+'
RE_TIMES_SIGN = r'\*'
RE_OPEN_BRACKET = r'\['
RE_CLOSE_BRACKET = r'\]'
RE_COLON = r':'

# Spacing and newline characters
RE_SPACING = r'[, \t.]+'
RE_NEWLINE = r'[|\n]'

# Extra information given by disassemblers
# We could just assume there will only ever be one set of disassembler information per line, and greedily match from
#   every open bracket '<' to the furthest end bracket '>', but I'm not sure this will always be the case.
# We first check for any string literals before looking at end brackets '>', in case those appear within the strings.
# So, we have the following restrictions on how disassembler information may be formatted:
#   1. All info is within '<>' brackets
#   2. You may nest '>' brackets within disassembler info, so long as they are always opened previously, and can go up to
#       a max depth of _DIS_INFO_MAX_REC_DEPTH
#   3. You may have any number of '<' brackets
#   4. Any '<>' brackets within well-formatted strings inside the disassembler info will have no effect as those are
#       treated as plain strings within the disassembler info
# This means the following disassembler infos are valid:
#   * <>
#   * <info>
#   * <lots of <<<<<<< things>
#   * <string="this is <> weird >>>>> string <<<<<">
#   * <nested <brackets> within data <are <ok> up to a certain <depth>>
# While the following are invalid and will break things:
#   * <
#   * <data>>
#   * <super<deep<nested<...<thing>>...>>
# NOTE: this recursive depth restriction exists because 1. I can't find any way of doing recursive regex's with python's
#   re module and 2. They'd probably be slower even if they did exist since we probably don't need that much depth
# NOTE: currently, the largest depth I've seen is 1 from something like: "invokespecial 0x0001<java/lang/object::<init>>"
DISINFO_START = '<'
DISINFO_END = '>'
_DIS_INFO_MAX_REC_DEPTH = 2  # The depth is 0 for no nested, +1 for each depth of nesting. EG: 2 would match up to "<0 <1 <2> 1> 0>"
RE_DISASSEMBLER_INFO = reduce(lambda rs, s: rs % s, [r'{ds}(?:{strlit}|%s|[^{de}])*{de}'] * _DIS_INFO_MAX_REC_DEPTH + [r'{ds}(?:{strlit}|[^{de}])*{de}'])\
    .format(strlit=RE_STRING_LITERAL, ds=DISINFO_START, de=DISINFO_END)


_RAISE_ERR = object()
def imm_to_int(token, on_err=_RAISE_ERR):
    """Convert the given value to integer
    
    If token is an integer, returns token. Otherwise, converts a string token to an integer, then back to a string, 
        accounting for hexadecimal, decimal, octal, and binary values

    Args:
        token (Union[str, int]): the immediate token to convert to integer
        on_err (Optional[Any]): if passed, then this value will be returned if there is an error while trying to parse
            the immediate value. Otherwise the error will just be raised like normal

    Returns:
        int: integer value of given token
    """
    try:
        return int(token, 0) if isinstance(token, str) else token
    except Exception:
        if on_err is _RAISE_ERR:
            raise
        return on_err


def scan_for_token(token_list, type=None, token=None, stop_on_type=None, stop_on_token=None, ignore_type=None, ignore_token=None, 
                   stop_unmatched=False, match_re=False, ignore_re_case=True, start=0, increment=1, wrap=True, max_matches=1, 
                   ret_list=False, ret='index', on_no_match=None):
    """Scans the given token list looking for a specific token(s) or token type(s)

    Will return None if no match is found.

    Detects tokens in the order:

        1. 'ignore' tokens
        2. 'stop' tokens
        3. accepted tokens (from `type` or `token` parameters)
    
    So, if one passes multiple parameters that conflict with one another, the above ordering is what takes precedent.
    
    Args:
        token_list (List[Tuple[str, str, ...]]): the list of tokens. Each element should be a tuple of (token_type, token, ...).
            The first element is the type of the token, second is the string token, and anything else is ignored. This means
            this function can work with either the 2-tuple token lists from Tokenizer() objects as well as the 3-tuple
            token lists from Normalizer() objects.
        type (Optional[Union[str, Iterable[str]]]): the type or types of tokens to return. Can be a string to only return one
            type of token, or an iterable of strings to return the first token found that has any of those types. If 
            `token` is not None, then the returned token must also match that argument.
            NOTE: you can match "not X" by using python re's negative lookahead: r'(?![X]).*', where '[X]' is the thing to not match
        token (Optional[Union[str, Iterable[str]]]): the token to return. Can be a string to only return one matching token, or
            an iterable of strings to return the first token found that matches any of those tokens. If `type` is None, then
            the returned token must also match that type.
            NOTE: you can match "not X" by using python re's negative lookahead: r'(?![X]).*', where '[X]' is the thing to not match
        stop_on_type (Optional[Union[str, Iterable[str]]]): if a token of this type is found, then we immediately stop
            searching and return whatever we currently have. Can be a string to only stop at one type of token, or an 
            iterable of strings to stop at the first token found that has any of those types.
            NOTE: you can match "not X" by using python re's negative lookahead: r'(?![X]).*', where '[X]' is the thing to not match
        stop_on_token (Optional[Union[str, Iterable[str]]]): if this token is found, then we immediately stop
            searching and return whatever we currently have. Can be a string to only stop at one token, or an 
            iterable of strings to stop at the first token found that matches any of these.
            NOTE: you can match "not X" by using python re's negative lookahead: r'(?![X]).*', where '[X]' is the thing to not match
        ignore_type (Optional[Union[str, Iterable[str]]]): ignores token types. Can be a string to only ignore one token type, or an 
            iterable of strings to ignore any token types that match any of these. These tokens will not be added to return
            lists or considered tokens to keep. Since these are checked before 'stop' token types, this will override
            the stopping on any tokens also matched with `stop_on_type`.
            NOTE: you can match "not X" by using python re's negative lookahead: r'(?![X]).*', where '[X]' is the thing to not match
        ignore_token (Optional[Union[str, Iterable[str]]]): ignores tokens. Can be a string to only ignore one token, or an 
            iterable of strings to ignore any tokens that match any of these. These tokens will not be added to return
            lists or considered tokens to keep. Since these are checked before 'stop' token types, this will override
            the stopping on any tokens also matched with `stop_on_token`.
            NOTE: you can match "not X" by using python re's negative lookahead: r'(?![X]).*', where '[X]' is the thing to not match
        stop_unmatched (bool): if True, will stop on the first unmatched token. IE: a token that was not ignored, was not
            already stopped on, and was not considered a token to keep
        match_re (bool): if True, will assume any match values in `type` or `token` are to be considered regular expressions to fullmatch()
        ignore_re_case (bool): if True, will pass re.IGNORECASE as a flag when making the regular expressions
        start (int): the index to start at within token_list
        increment (int): the increment to use when searching for tokens. Set to a negative number to move backwards through the list
            NOTE: if returning multiple values, they will be returned in the order they appear in the input list, regardless
            of the `increment` value
        wrap (bool): if True, then the initial `start` index will be wrapped to the length of the `token_list`. If False,
            then an initial `start` index that is out of bounds of the `token_list` will immediately stop.
        max_matches (Union[int, None]): the number of matches to find. If 1, then values will be returned as normal. If >1, then this will
            search through the list finding up to `max_matches` matching tokens and return their `ret` values as a list in the
            order that they were found. If None, then all matches found will be returned
            NOTE: if `max_matches` != 1, then the return value will always either be None if no matches were found, or a list (even if
            only one match was found)
        ret_list (bool): if True, will always return a list, even if only a single return value was present
        ret (Union[str, Iterable[str]]): what value(s) to return. Can be a single string to return a single value, or an 
            iterable of strings to return multiple values as a tuple in the order they were passed. Valid strings:

            - 'index': return the index in `token_list` of the matched token
            - 'type': return the token type of the matched token
            - 'token': return the string token that was matched
            - 'all': return all of the above. If in a passed list, ignores all other values in the list. Will return values
              in the order above.
        
        on_no_match (Optional[Any]): value to return if there were no matches found. Defaults to None
    
    Returns:
        Union[None, int, str, Tuple, List]: None if no match is found, or one of the return types designated by `ret` argument,
            or a tuple of multiple return values if user passed multiple values in `ret`, or a list of one of the previous
            if collecting matches for multiple tokens.
            NOTE: if returning multiple values, they will be returned in the order they appear in the input list, regardless
            of the `increment` value
    """
    # Compile into RE's if using
    re_flags = re.IGNORECASE if ignore_re_case else 0
    _mre = lambda val: [(re.compile(t, flags=re_flags) if match_re and not isinstance(t, re.Pattern) else t) for t in 
                        ([] if val is None else [val] if isinstance(val, str) else list(val))]
    type, token, stop_on_type, stop_on_token, ignore_type, ignore_token = \
        map(_mre, [type, token, stop_on_type, stop_on_token, ignore_type, ignore_token])

    # Make sure the user passed at least one type or token
    if len(type) == 0 and len(token) == 0:
        raise ValueError("Must pass at least one `type` or `token` to match to")
    
    # Make sure some integer values are good
    if increment == 0:
        raise ValueError("`increment` cannot be 0")
    max_matches = 2**100000 if max_matches is None else max_matches
    if max_matches < 1:
        raise ValueError("`num_matches` must be >= 1")
    
    # Make sure the user passed a valid return value
    ret = [ret] if isinstance(ret, str) else list(ret)
    ret = [r.lower() for r in ret]
    for i, r in enumerate(ret):
        if r in ['ind', 'idx', 'index', 'loc']:
            ret[i] = 0
        elif r in ['type', 'token_type']:
            ret[i] = 1
        elif r in ['token', 'token_string']:
            ret[i] = 2
        elif r in ['all']:
            ret = list(range(3))
            break
        else:
            raise ValueError("Unknown return type: %s" % repr(r))
    
    # If the token_list is empty, return None. Otherwise, insert the indices into the token_list at the beginning
    if len(token_list) == 0:
        return None
    token_list = [[i, t[0], t[1]] for i, t in enumerate(token_list)]
    
    # Function to check string matches depending on whether it is a regular expression or just normal string
    _str_match = lambda _token, _match: _match.fullmatch(_token) is not None if isinstance(_match, re.Pattern) else (_token == _match)
    
    # Iterate through the token_list finding all matches until reaching the end
    idx = (start % len(token_list)) if wrap else start
    ret_inds = []
    while 0 <= idx < len(token_list):
        # Add in the increment now to make the code nicer
        curr_idx = idx
        idx += increment

        # Check if this element is one that should be ignored, and skip it if so
        if any(any(_str_match(token_list[curr_idx][i + 1], t) for t in arr) for i, arr in enumerate([ignore_type, ignore_token])):
            continue
        # Check if this element matches a stop type or token, and break if so
        elif any(any(_str_match(token_list[curr_idx][i + 1], t) for t in arr) for i, arr in enumerate([stop_on_type, stop_on_token])):
            break
        # Check if this element doesn't match either type or token, and skip it if so (or, break if stop_unmatched=True)
        elif any(len(arr) > 0 and not any(_str_match(token_list[curr_idx][i + 1], t) for t in arr) for i, arr in enumerate([type, token])):
            if stop_unmatched:
                break
            continue

        # Otherwise, this matches! Add it into the list, and check if we have found our max number of matches
        ret_inds.append(curr_idx)
        if len(ret_inds) >= max_matches:
            break
    
    # Now that we have all of the matched indices, make the return objects
    _make_ret = lambda i: token_list[i][ret[0]] if len(ret) == 1 else tuple(token_list[i][r] for r in ret)
    _rev = lambda l: list(reversed(l)) if increment < 0 else l
    ret = on_no_match if len(ret_inds) == 0 else _make_ret(ret_inds[0]) if max_matches == 1 else _rev([_make_ret(i) for i in ret_inds])

    return ret if isinstance(ret, list) or not ret_list else [ret]


def get_normalizer(normalizer):
    """Returns the normalizer being used.

    Args:
        normalizer (Union[str, Normalizer, type]): either a ``Normalizer`` object (IE: has a callable 'normalize' function), 
            or a string name of a built-in normalizer to use, or a type of a normalizer to instantiate with no args/kwargs
            passed. Accepted strings include: 'innereye', 'deepbindiff', 'safe', 'deepsemantic', 'unnormalized', 
            'compressed_stats', 'hpc_data'

    Raises:
        ValueError: for unknown string name of normalizer
        TypeError: if `normalizer` was not a string or ``Normalizer`` object

    Returns:
        Normalizer: a ``Normalizer`` object
    """
    
    if isinstance(normalizer, str):
        norm_str = normalizer.lower()

        # You can specify the opcode/instruction level tokenization
        tl_names = bincfg.TokenizationLevel.AUTO.value + bincfg.TokenizationLevel.INSTRUCTION.value + bincfg.TokenizationLevel.OPCODE.value
        matched = re.fullmatch(r'(.*)[\-_](%s)' % '|'.join(tl_names), norm_str)
        if matched is not None:
            tl = 'op' if matched.groups()[1] in bincfg.TokenizationLevel.OPCODE.value else \
                'inst' if matched.groups()[1] in bincfg.TokenizationLevel.INSTRUCTION.value else 'auto'
            norm_str = matched.groups()[0]
        else:
            tl = 'auto'

        if norm_str.endswith("_normalizer"):
            norm_str, *_ = norm_str.rpartition("_normalizer")
        elif norm_str.endswith("_norm"):
            norm_str, *_ = norm_str.rpartition("_norm")

        known_isa = None
        for s in ['x86', 'java']:
            if norm_str.startswith(s):
                _, known_isa, norm_str = norm_str.partition(s)
        
        if norm_str.startswith('_'):
            norm_str = norm_str[1:]
        
        def _check_isa(needed, allow_none=False):
            if not allow_none and known_isa is None:
                raise ValueError("Ambiguous normalizer string: %s, must pass ISA name in front (IE: 'x86_%s')" % (repr(norm_str), norm_str))
            
            if isinstance(needed, str):
                needed = (needed,)
            
            if known_isa is not None and known_isa not in needed:
                raise ValueError("Normalizer %s can only be used with %s ISA's, not %s" % (repr(norm_str), repr(needed), repr(known_isa)))

        
        if norm_str in ['innereye', 'inner', 'innereyenormalizer']:
            _check_isa('x86', allow_none=True)
            return bincfg.normalization.X86InnerEyeNormalizer(tokenization_level=tl)
        elif norm_str in ['deepbindiff', 'bindiff', 'deepbin', 'deepbindiffnormalizer']:
            _check_isa('x86', allow_none=True)
            return bincfg.normalization.X86DeepBinDiffNormalizer(tokenization_level=tl)
        elif norm_str in ['safe', 'safenormalizer']:
            _check_isa('x86', allow_none=True)
            return bincfg.normalization.X86SafeNormalizer(tokenization_level=tl)
        elif norm_str in ['deepsem', 'deepsemantic', 'semantic', 'deepsemanticnormalizer']:
            _check_isa('x86', allow_none=True)
            return bincfg.normalization.X86DeepSemanticNormalizer(tokenization_level=tl)
        
        elif norm_str in ['none', 'unnorm', 'unnormalized', 'base', 'basenormalizer']:
            _check_isa(('x86', 'java'), allow_none=False)
            return bincfg.normalization.X86BaseNormalizer(tokenization_level=tl) if known_isa == 'x86' else\
                bincfg.normalization.JavaBaseNormalizer(tokenization_level=tl)
        elif norm_str in ['compressed', 'stats', 'comp_stats', 'compressed_stats', 'statistics', 'compressedstats', 'compressedstatsnormalizer']:
            _check_isa(('x86', 'java'), allow_none=False)
            if known_isa == 'java':
                raise NotImplementedError("Need to implement the java compressed stats normalizer")
            return bincfg.normalization.X86CompressedStatsNormalizer(tokenization_level=tl) if known_isa == 'x86' else\
                bincfg.normalization.JavaCompressedStatsNormalizer(tokenization_level=tl)
        elif norm_str in ['hpc', 'hpc_data', 'hpcdata', 'hpcdatanormalizer', 'hpcdatanorm', 'hpcnorm', 'hpcnormalizer']:
            _check_isa(('x86', 'java'), allow_none=False)
            if known_isa == 'java':
                raise NotImplementedError("Need to implement the java hpcdata normalizer")
            return bincfg.normalization.X86HPCDataNormalizer(tokenization_level=tl) if known_isa == 'x86' else\
                bincfg.normalization.JavaHPCDataNormalizer(tokenization_level=tl)
        else:
            raise ValueError("Unknown normalization string: '%s'" % normalizer)
    
    elif isinstance(normalizer, type):
        try:
            return get_normalizer(normalizer())
        except Exception as e:
            raise ValueError("Could not build a default normalizer from type: %s\nError Message: %s\n Traceback:%s"
                             % (repr(normalizer.__name__), e, traceback.format_exc()))
    
    elif hasattr(normalizer, 'normalize') and callable(normalizer.normalize):
        return normalizer
    
    else:
        raise TypeError("Unknown normalizer type: '%s'" % normalizer)


def parse_disinfo_json(string):
    """Attempts to pase a JSON object inside of disassembler info tokens
    
    Assumes the `DISINFO_START` and `DISINFO_END` have already been removed from the string.

    Args:
        string (str): the string to attempt to parse into json
    
    Returns:
        Union[None, JSONObject]: returns the resulting JSON object, or None if the string could not be parsed as JSON
    """
    try:
        return json.loads(string)
    except Exception:
        return None



"""
Miscellaneous utility functions
"""

import re
import time
import string
import pickle
import dis
import numpy as np
from enum import Enum
from threading import Thread
from hashlib import sha256
from inspect import currentframe, signature, _empty as EmptyDefaultParam
from types import MethodType, FunctionType
from collections import OrderedDict
from copy import deepcopy


# The actual progressbar object, once it has been determined
_IMPORT_PROGRESSBAR = None


# Python opcodes to ignore when checking similarity of functions
# Needed because starting python 3.11, there is a new 'RESUME' opcode that is essentially a nop, but does some stuff
#   for CPython internal debugging/tracing
IGNORE_OPNAMES = ['RESUME', 'NOP', 'CACHE']


def get_smallest_np_dtype(val, signed=False):
    """Returns the smallest numpy integer dtype needed to store the given max value.

    Args:
        val (int): the largest magnitude (furthest from 0) integer value that we need to be able to store
        signed (bool, optional): if True, then use signed ints. Defaults to False.

    Raises:
        ValueError: if a bad value was passed, or if the value was too large to store in a known integer size

    Returns:
        np.dtype: the smallest integer dtype needed to store the given max value
    """
    val = abs(val) if signed else val
    if val < 0:
        raise ValueError("Val must be >0 if using unsigned values: %d" % val)
    
    for dtype in ([np.int8, np.int16, np.int32, np.int64] if signed else [np.uint8, np.uint16, np.uint32, np.uint64]):
        if val < np.iinfo(dtype).max and (not signed or -val > np.iinfo(dtype).min):
            return dtype

    raise ValueError("Could not find an appropriate size for given integer: %d" % val)


def scatter_nd_numpy(target, indices, values):
    """Sets the values at `indices` to `values` in numpy array `target`
    
    Shamelessly stolen from: https://stackoverflow.com/questions/46065873/how-to-do-scatter-and-gather-operations-in-numpy

    Args:
        target (np.ndarray): the target ndarray to modify
        indices (np.ndarray): n-d array (same ndim as target) of the indices to set values to
        values (np.ndarray): 1-d array of the values to set

    Returns:
        np.ndarray: the resultant array, modified inplace
    """
    indices = tuple(indices.reshape(-1, indices.shape[-1]).T)
    np.add.at(target, indices, values.ravel())  # type: ignore
    return target


def arg_array_split(length, sections, return_index=None, dtype=np.uint32):
    """Like np.array_split(), but returns the indices that one would split at

    This will always return `sections` sections, even if `sections` > length (in which case, any empty sections will
    come at the end). If `sections` does not perfectly divide `length`, then any extras will be front-loaded, one per
    split array as needed.

    NOTE: this code was modified from the numpy array_split() source
    
    Args:
        length (int): the length of the sequence to split
        sections (int): the number of sections to split into
        return_index (Optional[int]): if not None, then an int to determine which tuple of (start, end) indices to
            return (IE: if you were splitting an array into 10 sections, and passed return_index=3, this would return
            the tuple of (start, end) indicies for the 4th split array (since we start indexing at 0))
        dtype (np.dtype): the numpy dtype to use for the returned array
    
    Returns:
        Union[np.ndarray, Tuple[int, int]]: a numpy array of length `sections + 1` where the split array at index `i`
            would use the start/end endices `[returned_array[i]:returned_array[i+1]]`, unless return_index is not None,
            in which case a 2-tuple of the (start_idx, end_idx) will be returned
    """
    if sections <= 0:
        raise ValueError('Number of sections must be > 0. Got %d' % sections)
    if length < 0:
        raise ValueError("Length must be >= 0. Got %d" % length)
    if return_index is not None and (return_index < 0 or return_index >= sections):
        raise ValueError("return_index, if not None, must be a positive integer in the range [0, sections). Got "
            "sections=%d, return_index=%d" % (sections, return_index))

    # If sections > length, then we would return the array [0, 1, 2, 3, ..., length - 1, length, length, length, ...]
    # NOTE: this also handles the case where length == 0 since sections > 0 always
    if sections >= length:
        ret_arr = np.arange(sections + 1, dtype=dtype)
        ret_arr[length:] = length

    # Otherwise we can do the normal divmod method
    # NOTE: this also handles the case where length == sections
    else:
        num_per_section, extras = divmod(length, sections)
        section_sizes = [0] + extras * [num_per_section + 1] + (sections - extras) * [num_per_section]
        ret_arr = np.array(section_sizes, dtype=dtype).cumsum()
    
    if return_index is not None:
        return ret_arr[return_index], ret_arr[return_index + 1]
    
    return ret_arr


# Some object types
_SingletonObjects = [None, Ellipsis, NotImplemented]
_DictKeysType = type({}.keys())
_GeneratorType = type((x for x in [1]))

_MAX_STR_LEN = 1000

# Types that are all able to be checked against one another using default '==' equality check
_DUNDER_EQ_TYPES = (int, float, np.number, complex, bytes, bytearray, memoryview, str, range, type, set, frozenset, _DictKeysType)

# Keep track of the current kwargs being used in equal()
_CURR_EQUAL_KWARGS = None
_EQ_DEFAULT_STRICT_TYPES = object()
_EQ_DEFAULT_UNORDERED = object()
_EQ_DEFAULT_RAISE_ERR = object()


# Context manager to return _CURR_EQUAL_KWARGS back to expected
class _ReturnCurrEqualKwargs:
    def __init__(self, kwargs):
        self.kwargs = kwargs
    def __enter__(self):
        return self
    def __exit__(self, *args):
        global _CURR_EQUAL_KWARGS
        if self.kwargs['control_kwargs']:
            _CURR_EQUAL_KWARGS = None
        else:
            for k in self.kwargs:
                if k.startswith('prev_'):
                    _CURR_EQUAL_KWARGS[k[len('prev_'):]] = self.kwargs[k]


def eq_obj(a, b, selector=None, strict_types=_EQ_DEFAULT_STRICT_TYPES, unordered=_EQ_DEFAULT_UNORDERED, raise_err=_EQ_DEFAULT_RAISE_ERR):
    """
    Determines whether a == b, generalizing for more objects and capabilities than default __eq__() method.
    Equal() is an equivalence relation, and thus:
    
        1. equal(a, a) is always True                       (reflexivity)
        2. equal(a, b) implies equal(b, a)                  (symmetry)
        3. equal(a, b) and equal(b, c) implies equal(a, c)  (transitivity)
    
    NOTE: This method is not meant to be very fast. I will apply as many optimizations as feasibly possible that I can
    think of, but there will be various inefficient conversions of types to check equality.
    
    NOTE: kwargs passed to the initial :func:`~gstats_utils.pythonutils.equality.equal` function call will be passed to 
    all subcalls, including those done in other objects using their built-in __eq__ function. Any objects can override
    those kwargs for any later subcalls (but not those above/adjacent). 
    NOTE: The `selector` kwarg is only used once, then consumed for any later subcalls
    
    Args:
        a (Any): object to check equality
        b (Any): object to check equality
        selector (Optional[str]): if not None, then a string that determines the 'selector' to use on both objects for
            determining equality. It should start with either a letter (case-sensitive), underscore '_', dot '.' or
            bracket '['. This string will essentially be appended to each object to get some attribute to determine
            equality of instead of the objects themselves. For example, if you have two lists, but only want to check
            if their element at index '2' are equal, you could pass `selector='[2]'`. This is useful for debugging purposes
            as the error messages on unequal objects will be far more informative. Defaults to None.
            NOTE: if you pass a `selector` string that starts with an alphabetical character, it will be assumed to be
            an attribute, and this will check equality on `a.SELECTOR` and `b.SELECTOR`
        strict_types (bool): if True, then the types of both objects must exactly match. Otherwise objects which are 
            equal but of different types will be considered equal. Defaults to False.
        unordered (bool): if True, then all known sequential objects (list, tuple, numpy array, etc.) will be considered
            equal even if elements are in a different order (eg: a multiset equality). Otherwise, sequential objects are
            expected to have their subelements appear in the same order. If the passed objects are not sequential, then
            this has no effect. Defaults to False.
        raise_err (bool): if True, then an ``EqualityError`` will be raised whenever `a` and `b` are unequal, along with
            an informative stack trace as to why they were determined to be unequal. Defaults to False.
    
    Raises:
        EqualityError: if the two objects are not equal, and `raise_err=True`
        EqualityCheckingError: if there was an error raised during equality checking
    
    Returns:
        bool: True if the two objects are equal, False otherwise
    """
    # Check if we are the first call and thus should controll the kwargs
    global _CURR_EQUAL_KWARGS
    _stack_kwargs = {'control_kwargs': False}
    if _CURR_EQUAL_KWARGS is None:
        _stack_kwargs['control_kwargs'] = True
        _CURR_EQUAL_KWARGS = {'strict_types': False, 'unordered': False, 'raise_err': False}
    
    # Update the kwargs if needed, otherwise grab them from the curr kwargs
    _stack_kwargs.update({'prev_strict_types': _CURR_EQUAL_KWARGS['strict_types'], 'prev_unordered': _CURR_EQUAL_KWARGS['unordered'],
        'prev_raise_err': _CURR_EQUAL_KWARGS['raise_err']})
    if strict_types is not _EQ_DEFAULT_STRICT_TYPES:
        _CURR_EQUAL_KWARGS['strict_types'] = strict_types
    else:
        strict_types = _CURR_EQUAL_KWARGS['strict_types']
    
    if unordered is not _EQ_DEFAULT_UNORDERED:
        _CURR_EQUAL_KWARGS['unordered'] = unordered
    else:
        unordered = _CURR_EQUAL_KWARGS['unordered']

    if raise_err is not _EQ_DEFAULT_RAISE_ERR:
        _CURR_EQUAL_KWARGS['raise_err'] = raise_err
    else:
        raise_err = _CURR_EQUAL_KWARGS['raise_err']

    # Cover with a context manager to reset _CURR_EQUAL_KWARGS back to expected values
    with _ReturnCurrEqualKwargs(_stack_kwargs):
        
        # Get the right selector, raising an error if it's bad
        if selector is not None:
            if not isinstance(selector, str):
                raise TypeError("`selector` arg must be str, not %s" % repr(type(selector).__name__))
            if selector == '':
                selector = None
            elif selector[0].isalpha():
                selector = '.' + selector
            elif selector[0] not in '._[':
                raise ValueError("`selector` string must start with a '.', '_', '[', or alphabetic character: %s" % repr(selector))
        
        # Use `selector` if needed
        if selector is not None:
            try:
                _failed_obj_name = 'a'
                _check_a = eval('a' + selector)
                _failed_obj_name = 'b'
                _check_b = eval('b' + selector)
                _failed_obj_name = None

                return eq_obj(_check_a, _check_b, selector=None, strict_types=strict_types, unordered=unordered, raise_err=raise_err)
            except EqualityError:
                raise EqualityError(a, b, "Objects had different sub-objects using `selector` %s" % repr(selector))
            except Exception:
                if _failed_obj_name is None:
                    raise EqualityCheckingError("Could not determine equality between objects a and b using `selector` %s\na: %s\nb: %s" %
                        (repr(selector), _limit_str(a), _limit_str(b)))
                raise EqualityCheckingError("Could not use `selector` with value %s on object `%s`" % (repr(selector), _failed_obj_name))

        # Wrap everything in a try/catch in case there is an error, so it will be easier to spot
        try:

            # Do a quick first check for 'is' as they should always be equal, no matter what
            if a is b:
                return True
            
            # Check if there are strict types
            if strict_types and type(a) != type(b):
                return _eq_check(False, a, b, raise_err, message='Objects are of different types and `strict_types=True`.')
            
            ##################
            # Checking types #
            ##################

            # We already checked 'is', so this must be an error
            if any(a is x for x in _SingletonObjects) or isinstance(a, Enum):
                return _eq_check(False, a, b, raise_err)
            
            # Check for bool first that way int's and bool's cannot be equal
            elif isinstance(a, bool) or isinstance(b, bool):
                # Enforce that this is a bool no matter what. Bool's are NOT int's. I will die on this hill...
                if not _eq_enforce_types(bool, a, b, raise_err):
                    return False
                return _eq_check(a == b, a, b, raise_err, message=None)
            
            # Check for objects using '=='
            elif isinstance(a, _DUNDER_EQ_TYPES):
                if not _eq_enforce_types(_DUNDER_EQ_TYPES, a, b, raise_err):
                    return False
                return _eq_check(a == b, a, b, raise_err, message=None)
            
            # Check for sequences list/tuple
            elif isinstance(a, (list, tuple)):
                
                # Check that b is something that could be converted into a list/tuple nicely

                # If check_b is a numpy array, convert check_a to one and do a numpy comparison
                if isinstance(b, np.ndarray):
                    # Check if check_b is an object array, and if so, use lists, otherwise use numpy
                    if b.dtype == object:
                        return _check_with_conversion(a, None, b, list, unordered, raise_err, strict_types)
                    return _check_with_conversion(a, np.ndarray, b, None, unordered, raise_err)

                # Check for things to convert to list
                elif isinstance(b, (_GeneratorType, _DictKeysType)):
                    return _check_with_conversion(a, None, b, list, unordered, raise_err)
                
                # Otherwise, make sure check_b is a list/tuple
                elif not isinstance(b, (list, tuple)):
                    return _eq_check(False, a, b, raise_err, message="checked b type could not be converted into list/tuple")
                
                # This is where we handle the actual checking.
                # Check that they are the same length
                if len(a) != len(b):
                    return _eq_check(False, a, b, raise_err, message="Objects had different lengths: %d != %d" % (len(a), len(b)))
                
                # If we are using ordered, then we can just naively check, otherwise, we have to do some other things...
                if not unordered:
                    # Check each element in the lists
                    for i, (_checking_a, _checking_b) in enumerate(zip(a, b)):
                        try:
                            # It will have returned an error if raise_err, so just return False
                            if not eq_obj(_checking_a, _checking_b, selector=None, strict_types=strict_types, unordered=unordered, raise_err=raise_err):
                                return False
                        except EqualityError:  # If we get an equality error, then raise_err must be true
                            raise EqualityError(a, b, "Values at index %d were not equal" % i)
                        except Exception:
                            raise EqualityCheckingError("Could not determine equality between elements at index %d" % i)
                    
                    # Now we can return True
                    return True

                # Unordered list checking
                else:
                    raise NotImplementedError
            
            # Check for numpy array
            elif isinstance(a, np.ndarray):

                # Ensure the other value can be converted into an array
                if not isinstance(b, np.ndarray):
                    # If check_a is an object array, then just convert it to a list now and have that check it
                    if a.dtype == object:
                        return _check_with_conversion(a, list, b, None, unordered, raise_err, strict_types)
                    
                    # Otherwise, if it is a known convertible, convert it
                    if isinstance(b, (list, tuple, _GeneratorType)):
                        return _check_with_conversion(a, None, b, np.array, unordered, raise_err, strict_types)
                    
                    # Otherwise, assume not equal
                    return _eq_check(False, a, b, raise_err, message="Could not convert b object of type %s to numpy array" % type(b).__name__)

                # Check if we are using objects or a different dtype
                if a.dtype == object:
                    # Attempt to check using lists at this point
                    return _check_with_conversion(a, list, b, list, unordered, raise_err, strict_types)

                # Otherwise, check if we are doing unordered or ordered.
                if not unordered:
                    # we can use the builtin numpy assert equal thing
                    try:
                        np.testing.assert_equal(a, b)
                        return True
                    except AssertionError as e:
                        return _eq_check(False, a, b, raise_err, message='Numpy assert_equal found discrepancies:\n%s' % e)
                
                # Otherwise we need to do an unordered equality check. Just convert to a list at this point and check it
                else:
                    return _check_with_conversion(a, list, b, list, unordered, raise_err, strict_types)
            
            # Check for dictionaries
            elif isinstance(a, dict):
                # b must be a dictionary
                if not _eq_enforce_types(dict, a, b, raise_err, message='Dictionaries must be same type to compare'):
                    return False
                
                # Check all the keys are the same
                try:
                    if not eq_obj(a.keys(), b.keys(), selector=None, strict_types=strict_types, unordered=unordered, raise_err=raise_err):
                        return False
                except EqualityError:  # If we get an equality error, then raise_err must be true
                    a_un = set(k for k in a if k not in b)
                    b_un = set(k for k in b if k not in a)
                    raise EqualityError(a, b, message="Dictionaries had different .keys()\n`a`-unique keys: %s\n`b`-unique keys: %s" 
                                        % (_limit_str(a_un), _limit_str(b_un)))
                except Exception:
                    raise EqualityCheckingError("Could not determine equality between dictionary keys\na: %s\nb: %s" %
                        (_limit_str(a.keys()), _limit_str(b.keys())))
                
                # Check all the values are the same
                for k in a:
                    try:
                        if not eq_obj(a[k], b[k], selector=None, strict_types=strict_types, unordered=unordered, raise_err=raise_err):
                            return False
                    except EqualityError:  # If we get an equality error, then raise_err must be true
                        raise EqualityError(a, b, message="Values at key %s differ" % repr(k))
                    except Exception:
                        raise EqualityCheckingError("Could not determine equality between dictionary values at key %s" % repr(k))
                
                # Now we can return True
                return True
            
            elif isinstance(a, (FunctionType, MethodType)):
                if not _eq_enforce_types((FunctionType, MethodType), a, b, raise_err, message='Functions must be same type to compare'):
                    return False
                
                # NOTE: don't do a '==' check here cause in earlier python versions (EG: 3.7), it causes an infinite
                #   recursion. For some reason, the '==' check also checks if the outer object a method is a part of
                #   is also equal, but only on older python versions. On 3.9, it works just fine

                da = _get_function_bytecode(a)
                db = _get_function_bytecode(b)
                
                try:
                    return eq_obj(da, db, selector=None, strict_types=strict_types, unordered=unordered, raise_err=raise_err)
                except EqualityError:  # If we get an equality error, then raise_err must be true
                    raise EqualityError(a, b, message="Functions contain different bytecode:\nA: %s\nB: %s" % (da, db))
                except Exception as e:
                    raise EqualityCheckingError("Could not determine equality between functions\nFor Reason: %s" % str(e))
            
            # Otherwise, use the default equality measure
            else:
                try:
                    return _eq_check(a == b, a, b, raise_err, message='Using built-in __eq__ equality measure')
                except EqualityError:  # If we get an equality error, then raise_err must be true
                    raise EqualityError(a, b, message="Values were not equal using built-in __eq__ method")
                except Exception as e:
                    raise EqualityCheckingError("Could not determine equality between dictionary values using built-in __eq__ method\nFor Reason: %s" % str(e))
        
        except EqualityError:
            raise
        except Exception as e:
            raise EqualityCheckingError("Could not determine equality between objects\na: %s\nb: %s\nFor reason: %s" % (_limit_str(a), _limit_str(b), str(e)))


def _check_with_conversion(a, type_a, b, type_b, unordered, raise_err, strict_types=False):
    """Attempts to convert check_a into type_a and check_b into type_b (by calling the types), then check equality on those
    
    Gives better error messages when things go wrong. You can pass None to one of the types to not change type. Pass the
    type itself (instead of a function) for better nameing on error messages about what they were being converted into.
    The name is given by type_a.__name__ if type_a is a type, or 'a lambda function' if it is an annonymus function, or
    the module + function name if a function
    """
    ca_type, check_a_str = _get_check_type(type_a)
    cb_type, check_b_str = _get_check_type(type_b)

    conversion_str = ('(with a value being converted using %s and b value being converted using %s)' % (check_a_str, check_b_str))\
            if check_a_str and check_b_str else \
        ('(with a value being converted using %s)' % check_a_str) if check_a_str else \
        ('(with b value being converted using %s)' % check_b_str) if check_b_str else \
        ''

    try:
        return eq_obj(ca_type(a), cb_type(b), selector=None, strict_types=strict_types, unordered=unordered, raise_err=raise_err)
    except EqualityCheckingError:
        raise
    except Exception:
        _eq_check(False, a, b, raise_err, message="Values were not equal %s" % conversion_str)


def _get_check_type(t):
    """Returns a function to call and a string describing what is being used to convert type given the type to convert
    
    Returns a tuple of (conversion_callable, type_description_string). The string will be empty if the conversion is
    the identity, t.__name__ if t is a type, 'a lambda function' if it is an anonymous function, or the module + 
    function/class name if it is a callable.
    """
    if t is None:
        return lambda x: x, ''

    if not callable(t):
        raise EqualityCheckingError("Cannot convert object types as given `type` is not callable: %s" % repr(t))
    
    return t, ('type ' + repr(t.__name__)) if isinstance(t, type) else repr(t)


def _eq_enforce_types(types, a, b, raise_err, message=None):
    """enforces check_b is of the given types using isinstance"""
    if not isinstance(a, types) or not isinstance(b, types):
        return _eq_check(False, a, b, raise_err, 'Objects were of incompatible types. %s' % message)
    return True


def _eq_check(checked, a, b, raise_err, message=None):
    """bool equal check, determine whether or not we need to raise an error with info, or just return true/false"""
    if not checked:
        if raise_err:
            raise EqualityError(a, b, message)
        return False
    return True


def _get_function_bytecode(func):
    """Returns a list of (OPNAME, ARG) tuples, normalized as much as I care to"""
    bytecode = [(inst.opname, inst.arg) for inst in dis.Bytecode(func) if inst.opname not in IGNORE_OPNAMES]

    # This is really just so our tests pass in 3.12...
    ret = []
    for opname, arg in bytecode:
        if opname == 'RETURN_CONST':
            ret += [('LOAD_CONST', arg), ('RETURN_VALUE', None)]
        else:
            ret.append((opname, arg))
    
    return ret


class _TimeoutFuncThread(Thread):
    """
    A simple Thread class to call the passed function with passed args/kwargs
    """
    def __init__(self, func, *args, **kwargs):
        """
        :param func: the function to call
        :param args: *args to pass to function when calling
        :param kwargs: **kwargs to pass to function when calling
        """
        super().__init__()
        self._func, self._args, self._kwargs = func, args, kwargs
        self._return = None
    
    def run(self):
        """
        This should never be called. Instead, call TimeoutFuncThread.start() to start thread
        """
        self._return = self._func(*self._args, **self._kwargs)


def timeout_wrapper(timeout=3, timeout_ret_val=None):
    """
    Wraps a function to allow for timing-out after the specified time. If the function has not completed after timeout
        seconds, then the function will be terminated.
    """
    def decorator(func):
        def wraped_func(*args, **kwargs):
            thread = _TimeoutFuncThread(func, *args, **kwargs)
            thread.start()

            init_time = time.time()
            sleep_time = 1e-8
            while time.time() - init_time < timeout:
                if thread.is_alive():
                    time.sleep(sleep_time)
                    sleep_time = min(0.1, sleep_time * 1.05)
                else:
                    return thread._return
            
            # If we make it here, there is an error, return value
            return timeout_ret_val
    
        return wraped_func
    return decorator


# Fail if string conversion takes > 10 seconds
_STR_CONV_TIMEOUT_SECONDS = 10

@timeout_wrapper(timeout=_STR_CONV_TIMEOUT_SECONDS, timeout_ret_val="[ERROR: String conversion timed out. Max time: %d seconds]" % _STR_CONV_TIMEOUT_SECONDS)
def _limit_str(a, limit=_MAX_STR_LEN):
    a_str = repr(a)
    return a_str if len(a_str) < limit else (a_str[:limit] + '...')


class EqualityError(Exception):
    """Error raised whenever an :func:`~gstats_utils.pythonutils.equality.equal` check returns false and `raise_err=True`"""

    def __init__(self, a, b, message=None):
        message = "Values are not equal" if message is None else message
        super().__init__("Object a (%s) is not equal to object b (%s)\na: %s\nb: %s\nMessage: %s" % \
            (repr(type(a).__name__), repr(type(b).__name__), _limit_str(a), _limit_str(b), message))


class EqualityCheckingError(Exception):
    """Error raised whenever there is an unexpected problem attempting to check equality between two objects"""


def eq_obj_err(obj1, obj2):
    """Same as eq_obj, but always raises an error"""
    return eq_obj(obj1, obj2, raise_err=True)


def hash_obj(obj, return_int=False):
    """Hashes the given object

    Args:
        obj (Any): the object to hash
        return_int (bool, optional): by default this method returns a hex string, but setting return_int=True will 
            return an integer instead. Defaults to False.

    Returns:
        Union[str, int]: hash of the given object
    """
    string = ""
    if obj is None:
        string += '[None]'
    elif isinstance(obj, (str, bool)):
        string += '(' + type(obj).__name__ + ') ' + str(obj)
    elif isinstance(obj, (int, np.integer)):
        string += '(int) ' + str(obj)
    elif isinstance(obj, (float, np.floating)):
        string += '(float) ' + str(obj)
    elif isinstance(obj, (list, tuple)):
        string += '(' + type(obj).__name__ + ') '
        for o in obj:
            string += hash_obj(o)
    elif isinstance(obj, (set, frozenset)):
        string += '(' + type(obj).__name__ + ') ' + str(sum(hash_obj(o, return_int=True) for o in obj))
    elif isinstance(obj, dict):
        string += '(' + type(obj).__name__ + ') '
        string += str(sum(hash_obj(hash_obj(k) + ', ' + hash_obj(v), return_int=True) for k, v in obj.items()))
    elif isinstance(obj, np.ndarray):
        string += '(' + type(obj).__name__ + ') '
        if obj.dtype == object:
            for a in obj:
                string += hash_obj(a) + ' '
        else:
            string += str(obj.data.tobytes())
    elif isinstance(obj, re.Pattern):
        string += '(' + type(obj).__name__ + ') ' + hash_obj(obj.pattern)
    elif isinstance(obj, _DictKeysType):
        string += '(' + type(obj).__name__ + ') ' + hash_obj(list(obj))
    elif isinstance(obj, (MethodType, FunctionType)):
        string += '(' + type(obj).__name__ + ') ' + repr(_get_function_bytecode(obj))
    else:
        string += str(hash(obj))
    
    hasher = sha256()
    hasher.update(string.encode('utf-8'))
    return int(hasher.hexdigest(), 16) if return_int else hasher.hexdigest()
    


def get_module(package, raise_err=True, err_message=''):
    """Checks that the given package is installed, returning it, and raising an error if not

    Args:
        package (str): string name of the package
        raise_err (bool, optional): by default, this will raise an error if attempting to load the module and it doesn't 
            exist. If False, then None will be returned instead if it doesn't exist. Defaults to True.
        err_message (str): an error message to add on to any import errors raised

    Raises:
        ImportError: if the package cannot be found, and `raise_err=True`

    Returns:
        Union[ModuleType, None]: the package
    """
    try:
        import importlib
        return importlib.import_module(package)
    except ImportError:
        if raise_err:
            raise ImportError("Could not find `%s` package.%s" % (package, err_message))
        return None


def isinstance_with_iterables(obj, types, recursive=False, ret_list=False):
    """Checks that obj is one of the given types, allowing for iterables of these types

    Args:
        obj (Any): the obj to test type
        types (Union[type, Tuple[type, ...]]): either a type, or tuple of types that obj can be
        recursive (bool, optional): by default, this method will only allow iterables to contain objects of a type in 
            `types`. If `recursive=True`, then this will accept arbitrary-depth iterables of types in `types`. 
            Defaults to False.
        ret_list (bool, optional): if True, will return a single list of all elements (or None if the isinstance check 
            fails). Defaults to False.

    Returns:
        Union[List[Any], bool, None]: the return value
    """
    if isinstance(obj, types):
        return [obj] if ret_list else True
    
    try:
        if ret_list:
            ret = []
            for elem in obj:
                ret += [elem] if isinstance(elem, types) else isinstance_with_iterables(elem, types, recursive=True, ret_list=True) if recursive else None
            return ret
        else:
            for elem in obj:
                if not (isinstance(elem, types) or (recursive and isinstance_with_iterables(elem, types, recursive=True, ret_list=False))):
                    return False
    except:
        return None if ret_list else False
    

def paramspec_name(obj, file_ext=None, savedparam_funcname=None, valid_filename=None):
    """Returns a string name for the given object based on save paramspec info
    
    Requires that the @parameter_saver function decorator was used on at least one function on the given object and
    was called at least once.

    Args:
        obj (Any): the object to get the string name from
        file_ext (Optional[str]): optional file extension to add to the end of the returned string. A period '.' will be
            inserted between the paramspec name and the file_ext if it is not already present at the beginning of file_ext
        savedparam_funcname (Optional[str]): the name of the function to use to generate the paramspec name. If None, then
            it will default first to '__init__' if it exists, then to the first saved paramspec attached to the object
            (in order of when the functions were called). Otherwise, should be a string name of the function to use
        valid_filename (Optional[bool]): if True, then the returned string will be modified so that it works as a valid
            filename. If False, then no such transformation will be applied. Otherwise if None, then this will be True
            if file_ext is not None and False otherwise.
    """
    valid_filename = file_ext is not None if valid_filename is None else valid_filename

    # Get the paramspec we will be using
    if not hasattr(obj, '__savedparams__') or len(obj.__savedparams__) == 0:
        raise ValueError("Could not find the '__savedparams__' attribute on the given object, or it was empty. Was a function decorated with @parameter_saver called yet?")
    elif savedparam_funcname is None:
        savedparam = obj.__savedparams__['__init__'] if '__init__' in obj.__savedparams__ else list(obj.__savedparams__.values())[0]
    elif savedparam_funcname not in obj.__savedparams__:
        raise ValueError("Could not find passed savedparam_funcname (%s) in object's __savedparams__" % repr(savedparam_funcname))
    else:
        savedparam = obj.__savedparams__[savedparam_funcname]

    # Get all of the parameters in order, in case we need to go by naming
    all_params = OrderedDict()
    for param_dict in [savedparam['args'], savedparam['kwargs']]:
        for key, val in param_dict.items():
            all_params[key] = val
    
    # Convert all of the args/kwargs to an appropriate string name
    strings = [type(obj).__name__]
    for key in savedparam['naming']:
        strings.append(_clean_paramspec_str(all_params[key], valid_filename=valid_filename))
    
    # Combine the strings together
    ret = '_'.join(strings)

    # Add on the file extension if needed
    file_ext = None if file_ext is None else file_ext if file_ext.startswith('.') else ('.' + file_ext)
    return ret if file_ext is None else (ret + file_ext)


_CLEAN_PARAMSPEC_VALID_CHARS = set(string.ascii_letters + string.digits + "-_(){}.")
def _clean_paramspec_str(val, valid_filename=False):
    """Converts the given value to a string and cleans the string
    
    This will:
        - convert 'val' to a string
        - if valid_filename is True:
          * ensures all characters are either alphanumeric or in "-_(){}." - any characters that do not fit those rules
            will be replaced with '%xHH' where HH is the hex code of the character, or '%uHHHH' where HHHH is the hex
            code of the unicode character. The '%' character will be replaced with its hexcode like so: '%x25'
          * Any spaces will be removed, with no replacement
    """
    ret_str = ""
    val_str = ('{' + val.__paramspec_name__() + '}') if hasattr(val, '__paramspec_name__') else str(val)

    if valid_filename:
        for c in val_str:
            if c not in _CLEAN_PARAMSPEC_VALID_CHARS:
                ret_str += ('%%x%02x' % ord(c)) if ord(c) <= 255 else ('%%u%04x' % ord(c))
            else:
                ret_str += c
    else:
        ret_str = val_str

    return ret_str.replace(' ', '')


def parameter_saver(func=None, naming=None, not_naming=None, ignore=None, not_ignore=None, insert_functions=False, copy=True):
    """A function that can wrap object methods to save calls to those methods
    
    Should only be used on __init__, or some other function which is only called once in that object's lifecycle.

    Can be used both like:

    .. code-block:: python

        @parameter_saver
        def __init__(self, *args, **kwargs):
            ...
    
    or like:

    .. code-block:: python

        @parameter_saver()
        def __init__(self, *args, **kwargs):
            ...
    
    Subsequent calls to wrapped functions will not have their parameters saved.

    Adds two new attributes: '__savedparams__' and '__paramspec_name__':

        - '__savedparams__': a dictionary that has keys being the function names that this wrapper was applied to
          (EG: '__init__'), and values being a subdictionary with keys/values:

          * 'args' (OrderedDict[str, Any]): args that were passed on function call, in order with their argument names
          * 'kwargs' (OrderedDict[str, Any]): kwargs that were passed on function call, in order. NOTE: any extra args 
            that would spill over into kwargs will be saved here 
          * 'naming' (Set[str]): set of strings for parameters that will be used when calling paramspec_name()
          * 'ignore' (Set[str]): set of strings for parameters to ignore all together

    Args:
        func (Callable): the function to wrap, or None if we should return a function that will later wrap another function
        naming (Optional[Iterable[str]]): iterable of strings for which parameters should be used for naming. Only the 
            parameters with these names will be used when generating a name with paramspec_name() or obj.__paramspec_name__, 
            and they will be used in the order that they appear here. Default (None) is to use all parameters in the order 
            that they appear in the method signature. Mutually exclusive with `not_naming`
        not_naming (Optional[Iterable[str]]): iterable of strings for which parameters should NOT be used for naming. All
            other parameters will be used. Mutually exclusive with `naming`
        ignore (Optional[Iterable[str]]): iterable of strings for which parameters should be ignored. These parameters
            do not appear when calling paramspec_name() and will not be saved. Default (None) is to not ignore any
            parameters. Mutually exclusive with `not_ignore`
            NOTE: only keyword arguments can be ignored
        not_ignore (Optional[Iterable[str]]): iterable of strings for which parameters should NOT be ignored. All other
            parameters will be used. Mutually exclusive with `ignore`
            NOTE: only keyword arguments can be ignored
        insert_functions (bool): if True, then extra functions will be added to the object. This will add:
            
            * .save(path: str) function - pickles the object and saves it to the given path
            * .load(path: str) function - Adds this function at the class level. Attempts to load and return a pickled
              object from the given path, checking to make sure it is the correct type
            * __setstate__(state) function - re-initializes this object with the given state information. This will attempt
              to initialize the new object with __init__ and using the args/kwargs present in __savedparams__['__init__']
              if present, then will fill in the rest of the __dict__ attributes as normal

        copy (Union[bool, str]): if True, will attempt to copy parameters by checking if they have a `.copy()` method
            and calling it if so to produce the object that is saved, that way any updates to objects during/after
            initialization will not affect the saved parameters. If False, then the original object will be used. Can
            also be the string 'deep' to perform a deep copy of each object.
    """
    # Do some error checking and cleaning
    if naming is not None and not_naming is not None:
        raise ValueError("Cannot pass both 'naming' and 'not_naming' to parameter saver")
    if ignore is not None and not_ignore is not None:
        raise ValueError("Cannot pass both 'ignore' and 'not_ignore' to parameter saver")
    if copy not in [True, False, 'deep']:
        raise ValueError("`copy` keyword must be True, False, or 'deep', not: %s" % repr(copy))
    
    def _clean(_iter):
        if _iter is None:
            return _iter
        name = [k for k, v in currentframe().f_back.f_locals.items() if _iter is v][0]  # Now we're thinking with portals
        _iter = [_iter] if isinstance(_iter, str) else list(_iter)
        for elem in _iter:
            if not isinstance(elem, str):
                raise TypeError("Elements in %s must all be strings, not: %s" % (repr(name), repr(type(elem).__name__)))
        return _iter
    
    naming, not_naming, ignore, not_ignore = _clean(naming), _clean(not_naming), _clean(ignore), _clean(not_ignore)

    def _is_positional(param):
        return param.default is EmptyDefaultParam
    def _is_kwarg(param):
        return not _is_positional(param)

    def new_obj_func(self, *args, **kwargs):
        """Fun fact: we don't have to do too much error checking here since we immediately call func() after, and any
           errors with args, positional/keyword-only arguments, passing multiple values for same parameter name, etc.
           will be caught by python when we call it
        """
        # Need access to the outer scope 'func' since it may be later modified by 'wrapped_func()'
        nonlocal func

        # Get the signature, and make sure all values in naming/not_naming/ignore/not_ignore are correct parameters
        sig = signature(func).parameters
        for _iter in [naming, not_naming, ignore, not_ignore]:
            if _iter is None:
                continue
            for elem in _iter:
                if elem not in sig:
                    raise ValueError("Unknown parameter: %s" % repr(elem))

        # Get the function signature (only keeping those names which are not to be ignored)
        # Pad args so we can zip them with paramspec later, just in case we have any args that spill over into kwargs
        # Also remove the 'self' argument, which is always the first one
        paramspec = OrderedDict()
        ret_ignore = set()
        rm_args_inds = []
        for i, (k, v) in enumerate(list(sig.items())[1:]):
            # We ignore anything in 'ignore' (while checking that it isn't an arg), or anything that is NOT in not_ignore
            #   (but only if it is a kwarg, we don't raise an error here if it is an arg - just add the arg silently)
            if (ignore is not None and k in ignore) or \
                (not_ignore is not None and k not in not_ignore and _is_kwarg(v)):
                if _is_positional(v):
                    raise ValueError("Cannot ignore positional argument: %s" % repr(k))
                
                # Also check if this is in naming (but don't worry about not_naming, that will just use everything leftover)
                if naming is not None and k in naming:
                    raise ValueError("Cannot ignore parameter %s as it is also in 'naming'" % repr(k))

                ret_ignore.add(k)
                rm_args_inds.append(i)
            
            # Otherwise we keep the parameter
            else:
                paramspec[k] = v
        
        # Determine which parameters will be used in naming
        ret_naming = list(paramspec.keys()) if (naming is None and not_naming is None) else list(naming) if naming is not None else \
            list(k for k in paramspec.keys() if k not in not_naming)
        
        # Pad the args so it zips correctly later, allowing us to make use of extra args that spill into kwargs
        # Also remove any args that would spill over into ignored kwargs
        _ARG_Padder = object()
        cleaned_args = tuple(a for i, a in enumerate(args) if i not in rm_args_inds)
        padded_args = cleaned_args + (_ARG_Padder,) * max(0, len(paramspec) - len(cleaned_args))

        def _copy(obj):
            # Copy the arg if using
            if copy == 'deep':
                return deepcopy(obj)
            elif copy and hasattr(obj, 'copy') and callable(obj.copy):
                return obj.copy()
            return obj

        # Use an OrderedDict so parameters are always in order. Order to choose kwargs is: arg from 'args', kwarg from
        #   'kwargs', default value
        ret_kwargs = OrderedDict()
        ret_args = OrderedDict()
        for arg, (key, param_obj) in zip(padded_args, paramspec.items()):
                
            # If this is an arg in the function definition
            if _is_positional(param_obj):
                ret_args[key] = _copy(arg)

            # If this is a kwarg in the function definition
            else:
                ret_kwargs[key] = _copy(arg if arg is not _ARG_Padder else kwargs[key] if key in kwargs else param_obj.default)

        # Check to make sure the __savedparams__ attribute is set on the object
        if not hasattr(self, '__savedparams__'):
            self.__savedparams__ = {}

        if func.__name__ not in self.__savedparams__:
            self.__savedparams__[func.__name__] = {'args': ret_args, 'kwargs': ret_kwargs, 'naming': ret_naming, 'ignore': ret_ignore}
        self.__paramspec_name__ = MethodType(paramspec_name, self)

        if insert_functions:
            self.save = MethodType(_paramspec_save, self)
            if not hasattr(type(self), 'load'):
                type(self).load = classmethod(_paramspec_load)
            self.__setstate__ = MethodType(_paramspec_setstate, self)
            self.__getstate__ = MethodType(_paramspec_getstate, self)
        
        return func(self, *args, **kwargs)
    
    def wrapped_func(_func):
        nonlocal func
        func = _func
        return new_obj_func
    
    return wrapped_func(func) if func is not None else wrapped_func


# Functions that will be added to paramspec objects if using
def _paramspec_save(self, path):
    with open(path, 'wb') as f:
        pickle.dump(self, f)
def _paramspec_load(cls, path):
    with open(path, 'rb') as f:
        return pickle.load(f)
_PARAMSPEC_SELF_CALLS = set()
def _paramspec_getstate(self):
    """Only keep the objects which can be pickled/unpickled, and assume everything else will be handled by __init__"""
    # We need to make sure we don't do infinite recursive calls to this function. If we do, just don't pickle it
    global _PARAMSPEC_SELF_CALLS
    if id(self) in _PARAMSPEC_SELF_CALLS:
        raise AttributeError
    _PARAMSPEC_SELF_CALLS.add(id(self))

    ret = {}
    for k, v in self.__dict__.items():
        if k == '__paramspec_name__':
            continue

        try:
            x = pickle.loads(pickle.dumps(v))
        except AttributeError:
            continue
        except Exception as e:
            print("Object cannot be pickled/unpickled: %s\nDue to Error: %s\n" % (v, e))
            raise e
        ret[k] = v
    
    _PARAMSPEC_SELF_CALLS.remove(id(self))
    return ret
def _paramspec_setstate(self, state):
    """Expects state to be a dictionary
    
    Will attempt to initialize self with __savedparams__['__init__'] if present, then will fill in the rest of the
    __dict__ info as normal
    """
    if '__savedparams__' in state and '__init__' in state['__savedparams__']:
        self.__init__(*state['__savedparams__']['__init__']['args'].values(), **state['__savedparams__']['__init__']['kwargs'])
    for k, v in state.items():
        setattr(self, k, v)


def paramspec_set_class_funcs(ret_cls):
    """Sets class functions for paramspec things on the given class"""
    ret_cls.__paramspec_name__ = paramspec_name
    ret_cls.save = _paramspec_save
    if not hasattr(ret_cls, 'load'):
        ret_cls.load = classmethod(_paramspec_load)
    ret_cls.__setstate__ = _paramspec_setstate
    ret_cls.__getstate__ = _paramspec_getstate
    return ret_cls


class ParameterSaver(type):
    """A metaclass used to add in parameter saving to the initialization function
    
    This allows you to wrap __init__ of a class without having to worry about blocking IDE's from seeing its args/kwargs,
    and will apply the parameter saving to all child classes as well. Will default to insert_functions=True
    """
    def __new__(cls, name, bases, dct):
        ret_cls = super().__new__(cls, name, bases, dct)
        ret_cls.__init__ = parameter_saver(ret_cls.__init__, insert_functions=False)
        return paramspec_set_class_funcs(ret_cls)
    

def split_by_metadata_key(metadata, set_splits, split_key, rng=None, subgroupings=None, final_sublist_size=1, eps=1e-8):
    """Splits data based on arbitrary keys in its metadata. Allows for subgroupings as well

    NOTE: This requires that all of the values for split_key in all metadata dictionaries (as well as those for any
    subgroupings being used) are hashable types.

    NOTE: make sure you include an 'INDEX' key in all of the metadata values if the order they appear in the metadata
    is not the order they should be interpreted to have in file. IE: if your 'INDEX' column in file does not match up
    with the index of datapoints within the file

    Args:
        metadata (List[Dict]): metadata for the data being split. A list of metadata dictionaries from all elements that could
            be loaded by the dataloader. If this has an 'INDEX' column, then that will be used to determine the 'indices'
            that are returned by this method. Otherwise, the indices will just be the order of datapoints as they appear.
            Assumes that if the 'INDEX' column is present in the first element, it will be present in all, and vice-versa
        set_splits (Dict[Any, float]): Dict mapping dataset name to float percent of the total dataset that should be 
            allocated to that dataset name. If an OrderedDict, then data will be assigned with priority to earlier datasets
            in the case of too few 'unique' datapoints (by `split_key`), or uneven class sizes. Otherwise, order is
            arbitrary.
        split_key (Optional[Any]): the metadata key to use to split data by. If None, will split just by the number of
            datapoints in metadata
        rng (Optional[Union[int, RNG]]): integer random state, or numpy RNG object to use for rng, or None to not randomly
            select elements and instead grab them in the order that they appear in metadata. This will gather elements
            first in order of the unique keys that appear, then in order of individual metadata elements.
        subgroupings (Optional[Iterable[Any]]): If None, then this will split normally by metadata key. Otherwise, this
            can be string/int or a list of subelements which will act as a key or keys in the metadata to subgroup
            data by. Each key will be grouped in order to apply 'subgroupings' to the data. For example, if you were
            to split by the 'problem_uid' key, then subgroup by the 'submission_id' key, this would return a list of
            lists of indices as the value for each set_split. The first list would be at the 'problem_uid' level where
            all indices with the same problem_uid would appear in the same outer list. Each sublist would contain all
            indices with the same 'submission_id' key value from those grouped into the outer 'problem_uid'-level list.
            Multiple subgrouping keys may be used at the same time to create deeper nested groupings. You may subgroup by
            the same key as the splitting key, which would ensure that, when loading data, all examples with the same value 
            for its splitting key would be prioritized to load together.

            NOTE: the current loading RNG implementation will randomly select subelements from each level of list deeper
            and deeper until reaching the final layer, at which time all values within that final list will be taken
            together. This means that if you were to say, split by 'problem_uid', and subgroup by both 'problem_uid'
            and 'submission_id' in order. You would then lose out on the prioritization of loading values with the
            same 'problem_uid' all together. To help with this, you may use the `final_sublist_size` argument which
            will make the final sublists contain that many 'unique' indices. In this the above example, it would ensure
            that there are `final_sublist_size` *unique* submission_id's within each final sublist, and that sublist
            would contain all indices with 1. a 'problem_uid' that is within that outer sublist and 2. a 'submission_id'
            that is within that inner sublist. This way, one could ensure the loading multiple examples from the same 
            problem_uid each selection, and make sure that all compilations of the same submission_id are loaded at
            the same time as well.
        final_sublist_size (int): the max size of the final sublist, in terms of number of 'unique' elements. See the 
            note above in subgroupings for more info. Only used if `subgroupings` is not None
        eps (float): small epsilon value to pass to split_list_by_sizes() using `set_splits`, see that func for more info
    
    Returns:
        Dict[Any, List[SplitIndElement]]: dictionary mapping each key in set_splits to its list of SplitIndElement
            objects. Each SplitIndElement can either be an integer index, or a list of SplitIndElement. This allows for
            nested groupings of elements to choose when loading data. 
    """
    rng = np.random.default_rng(seed=rng) if rng is not None and not isinstance(rng, np.random.Generator) else rng
    subgroupings = [subgroupings] if isinstance(subgroupings, (str, int, float, complex, bool)) else \
        [] if subgroupings is None else subgroupings
    
    # If we aren't splitting by a key at all, just split by length. RNG shuffle if needed
    if split_key is None:
        inds = list(range(len(metadata))) if 'INDEX' not in metadata[0] else [m['INDEX'] for m in metadata]
        if rng is not None: rng.shuffle(inds)
        return {k: vl for k, vl in zip(set_splits.keys(), split_list_by_sizes(inds, list(set_splits.values()), eps=eps))}

    # Group all of the indices based on their splitting metadata key
    # Wrap in try-catch in case user passes an unhashable type, not a list for `metadata`, etc
    try:
        split_mapping = OrderedDict()
        for i, md in enumerate(metadata):
            split_mapping.setdefault(md[split_key], []).append(i)
    except Exception as e:
        exc_type = type(e) if type(e) in [ValueError, TypeError, KeyError] else ValueError
        raise exc_type("Could not create inverse mapping for `metadata` dictionaries using `split_key` %s for reason:\n%s: %s"
                       % (repr(split_key), type(e).__name__, e))

    # Split up the unique split_key values based on the set_splits. Use RNG if `rng` is not None
    unique_vals = list(split_mapping.keys())
    if rng is not None: rng.shuffle(unique_vals)
    
    # Split up the unique values into all the sets and create a mapping from each value to its associated set
    # Then, grab sort all the indicies into their associated set (we already have them sorted by value in split_mapping)
    val_to_set = {v: k for k, vl in zip(set_splits.keys(), split_list_by_sizes(unique_vals, list(set_splits.values()), eps=eps)) for v in vl}
    set_inds = {k: [] for k in set_splits.keys()}
    for split_val, inds in split_mapping.items():
        set_inds[val_to_set[split_val]] += inds

    # Go through each group of elements applying subgroupings if needed, and randomizing final lists if doing that
    ret = {}
    for set_name, inds_list in set_inds.items():
        if len(subgroupings) > 0:
            inds_list = _apply_subgroupings(metadata, inds_list, subgroupings, final_sublist_size, rng)
        if rng is not None:
            rng.shuffle(inds_list)
        ret[set_name] = inds_list
    
    # Get the actual 'INDEX' values if present
    if 'INDEX' in metadata[0]:
        ret = {k: _rec_list_select_metadata_index(inds, metadata) for k, inds in ret.items()}
    return ret


def _rec_list_select_metadata_index(inds, metadata):
    """Returns a new list with the same structure as inds, just with each integer instead being converted to the correct
       'INDEX' value within the metadata"""
    return [metadata[i]['INDEX'] if isinstance(i, int) else _rec_list_select_metadata_index(i, metadata) for i in inds]


def _apply_subgroupings(metadata, set_inds, subgroupings, final_sublist_size, rng):
    """Applys subgroupings and whatnot to split_by_metadata_key indices
    
    Args:
        metadata (List[Dict]): list of dictionaries of metadata
        set_inds (List[int]): list of integer indices in metadata to subgroup
        subgroupings (List[Any]): list of keys in metadatas to group by. Will group by one key at a time recursively
        final_sublist_size (int): the number of 'unique' elements to have in the final sublist
        rng (Optional[RNG]): either None or numpy rng
    
    Returns:
        List[SetIndsElement]: list of integer set inds or other sublists of such elements
    """
    subgroup_key, subgroupings = subgroupings[0], subgroupings[1:]  # Need this, can't pop

    # Find all of the unique values for this subgroup, and which inds have those values
    try:
        subgroup_mapping = OrderedDict()
        for i in set_inds:
            subgroup_mapping.setdefault(metadata[i][subgroup_key], []).append(i)
    except Exception as e:
        exc_type = type(e) if type(e) in [ValueError, TypeError, KeyError] else ValueError
        raise exc_type("Could not create inverse mapping for `metadata` dictionaries using `subgroupings` key %s for reason:\n%s: %s"
                       % (repr(subgroup_key), type(e).__name__, e))

    # Find the unique subgroup_mapping values and RNG them if needed. This applies enough shuffling for this whole function
    unique_vals = list(subgroup_mapping.keys())
    if rng is not None: rng.shuffle(unique_vals)
    
    # Create a mapping from each unique value to a unique integer for the list it should go in. Then, create a bunch
    #   of lists to store all those values and insert all of the indices into their associated group
    unique_vals_mapping = {u: i for i, u in enumerate(unique_vals)}
    ret = [[] for _ in range(len(unique_vals))]
    for subgroup_val, inds in subgroup_mapping.items():
        ret[unique_vals_mapping[subgroup_val]] += inds
    
    # If there are more subgroupings, apply them. Otherwise, merge into final_sublist_size chunks if needed
    if len(subgroupings) > 0:
        ret = [_apply_subgroupings(metadata, l, subgroupings, final_sublist_size, rng) for l in ret]
    elif final_sublist_size > 1:
        merge_inds = np.array_split(np.arange(len(ret)), np.ceil(len(ret) / final_sublist_size))
        ret = [[v for aind in arr_inds for v in ret[aind]] for arr_inds in merge_inds]

    return ret

    
def split_list_by_sizes(l, sizes, eps=1e-8):
    """Splits the given list into len(sizes) different lists in order based on sizes
    
    Elements will be inserted into returned lists in order, prioritizing first having at least one element per list, then
    biasing any remaining elements into earlier lists.

    Args:
        l (Iterable[Any]): the list of elements to split
        sizes (Union[Iterable[float], Iterable[int]]): the different sizes to apply. Can either be an iterable of floats
            in which case each element is a percent of the total data to keep and all elements should be >=0 and <=1 and
            all elements should sum to 1. Or, can be an iterable of integers in which case all elements should be >=0
            and <= len(`l`) and all elements should sum to len(`l`)
        eps (float): the epsilon value used to determine if sum(`sizes`) (when `sizes` is a float) is equal to 1
    
    Returns:
        List[List[Any]]: a list of all sublists
    """
    l, sizes = list(l), list(sizes)

    if len(sizes) == 0:
        raise ValueError("Must pass at least one size in `sizes`")

    # Check for float vs int and get amount to have in each return list. If you pass only the int 1, then count is as 1.0
    if isinstance(sizes[0], float) or (isinstance(sizes[0], int) and sizes[0] == 1 and len(sizes) == 1):
        if abs(sum(sizes) - 1) > eps:
            raise ValueError("Sum of float `sizes` values must be 1.0, got: %f" % sum(sizes))
        
        # Split data by percent, and make sure we prioritize filling any empty lists
        sizes = [int(s * len(l)) for s in sizes]
        remaining = len(l) - sum(sizes)
        for i in range(len(sizes)):
            if remaining <= 0: break
            if sizes[i] == 0:
                sizes[i] += 1
                remaining -= 1
        
        # Any extra remaining can be inserted from start to end
        while remaining > 0:
            for i in range(len(sizes)):
                if remaining <= 0: break
                sizes[i] += 1
                remaining -= 1
        
    elif isinstance(sizes[0], (int, np.integer)):
        if sum(sizes) != len(l):
            raise ValueError("Sum of integer `sizes` values must be length of input list %d, got: %d" % (len(l), sum(sizes)))
    else:
        raise TypeError("Unknown `sizes` element type: %s" % repr(type(sizes[0]).__name__))
    
    # Now sizes should contain the number of elements to get for each list. Do a cumsum and gather all indices
    sums = np.concatenate(([0], np.cumsum(sizes)))
    return [l[sums[i]:sums[i+1]] for i in range(len(sizes))]


class _tqdm_like_iter:
    def __init__(self, iterable):
        self.iterable = iter(iterable)
    
    def __next__(self):
        return next(self.iterable)
    
    def __iter__(self):
        return self.iterable

    def update(self, num):
        next_val = None
        for i in range(num):
            next_val = next(self)
        return next_val


def _using_progressbar(iterable, *args, progress=True, **kwargs):
    """Allows one to call progressbar(iterable, progress) to determine use of progressbar automatically.
    
    Checks to see if we are in a python notebook or not to determine which progressbar we should use.
    Copied from: https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
    """
    if not progress:
        return _tqdm_like_iter(iterable)

    global _IMPORT_PROGRESSBAR
    if _IMPORT_PROGRESSBAR is None:
        try:
            _tqdm_import = get_module('tqdm')
            _IMPORT_PROGRESSBAR = lambda *args, **kwargs: _tqdm_import.tqdm(*args, **kwargs)
        except ImportError:
            print("Could not import tqdm!")
            _IMPORT_PROGRESSBAR = _tqdm_like_iter
    
    return _IMPORT_PROGRESSBAR(iterable, *args, **kwargs)

progressbar = _using_progressbar



from .misc_utils import get_smallest_np_dtype, progressbar, eq_obj, eq_obj_err, isinstance_with_iterables, scatter_nd_numpy, \
    hash_obj, EqualityError, get_module, arg_array_split, parameter_saver, paramspec_name, ParameterSaver, paramspec_set_class_funcs, \
    split_by_metadata_key, split_list_by_sizes
from .atomic_token_dict import AtomicTokenDict
from .cfg_utils import get_special_function_names, check_for_normalizer, get_address, update_memcfg_tokens, update_atomic_tokens



import re
from .base_normalizer import BaseNormalizer
from .base_tokenizer import BaseTokenizer, TokenizationLevel, Tokens, Architectures, TokenMismatchError, UnknownTokenError, get_architecture
from .normalize import normalize_cfg_data, get_normalizer
from .x86 import *
from .java import *
from .multi_normalizer import MultiNormalizer


# Regular expression matching all known return instructions, hopefully without interference
RETURN_INSTRUCTION_RE = re.compile(r'[ \t\n]*(?:(?:{x86})|(?:{java})).*'.format(
    x86=r'(?:lock )*(?:ret[nf]?)',  # Have to allow for lock instructions, as well as retn (return near) and retf (return far)
    java=r'ret|[ilfda]?return',  # 'ret' is deprecated since Java 7 or something, but we should leave it here for backwards compatibility
))



"""Class that can use multiple normalization methods"""

from .base_tokenizer import _USE_DEFAULT_NT
from ..utils import hash_obj


class MultiNormalizer:
    """A normalizer that can work with multiple sub-normalizers based on architecture
    
    This does not inheret from BaseNormalizer, and thus you cannot modify or call most normalizer functions from this
    normalizer itself. It essentially just acts as a wrapper around multiple different normalizers.

    Parameters
    ----------
    normalizers: `BaseNormalizer`
        One or more normalizers to use together. May only use one per architecture.
    """
    def __init__(self, *normalizers):
        self._arch_to_norm = {}

        for n in normalizers:
            if n.tokenizer.ARCHITECTURE is None:
                raise ValueError("Cannot have a None architecture for a MultiNormalizer `normalizer`: %s" % repr(str(n)))
            elif n.tokenizer.ARCHITECTURE in self._arch_to_norm:
                raise ValueError("Found multiple tokenizers for the same architecture: %s" % repr(n.tokenizer.ARCHITECTURE))
            self._arch_to_norm[n.tokenizer.ARCHITECTURE] = n
    
    def normalize(self, *strings, cfg=None, block=None, newline_tup=_USE_DEFAULT_NT, match_instruction_address=True):
        """Normalizes the given iterable of strings.

        Args:
            strings (str): arbitrary number of strings to normalize
            cfg (Union[CFG, MemCFG], optional): either a ``CFG`` or ``MemCFG`` object that these lines occur 
                in. Used for determining function calls to self, internal functions, and external functions. If not 
                passed, then these will not be used. Defaults to None.
            block (Union[CFGBasicBlock, int], optional): either a ``CFGBasicBlock`` or integer block_idx in a ``MemCFG``
                object. Used for determining function calls to self, internal functions, and external functions. If not 
                passed, then these will not be used. Defaults to None.
            newline_tup (Tuple[str, str], optional): the tuple to insert inbetween each passed string, or None to not 
                insert anything. Defaults to self.tokenizer.DEFAULT_NEWLINE_TUPLE
            match_instruction_address (bool, optional): if True, will match instruction addresses. If there is an immediate
                value at the start of a line (IE: start of a string in `strings`, or immediately after a Tokens.NEWLINE
                or Tokens.INSTRUCTION_START [ignoring any Tokens.SPACING]), then that token will be converted into a
                Tokens.INSTRUCTION_ADDRESS token. If there is a Tokens.COLON immediately after that token (again, ignoring
                any Tokens.SPACING), then that first Tokens.COLON match will be appended (along with any inbetween Tokens.SPACING)
                to that Tokens.INSTRUCTION_ADDRESS token. For example, using the x86 tokenization scheme:

                    - "0x1234: add rax rax" -> [(Tokens.INSTRUCTION_ADDRESS, '0x1234:'), ...]
                    - "  0x1234     : add rax rax" -> [(Tokens.SPACING, '  '), (Tokens.INSTRUCTION_ADDRESS, '0x1234     :'), ...]
                    - "0x1234 add rax rax" -> [(Tokens.INSTRUCTION_ADDRESS, '0x1234'), ...]
                
            kwargs (Any): extra kwargs to pass along to tokenization method, and to store in normalizer state

        Returns:
            List[str]: a list of normalized string instruction lines
        """
        if cfg is None:
            raise ValueError("Must pass the cfg being used when calling the MultiNormalizer")
        
        return self._arch_to_norm[cfg.architecture](*strings, cfg=cfg, block=block, newline_tup=newline_tup, 
                                                  match_instruction_address=match_instruction_address)
    
    def __call__(self, *strings, cfg=None, block=None, newline_tup=_USE_DEFAULT_NT, match_instruction_address=True):
        """Normalizes the given iterable of strings.

        Args:
            strings (str): arbitrary number of strings to normalize
            cfg (Union[CFG, MemCFG], optional): either a ``CFG`` or ``MemCFG`` object that these lines occur 
                in. Used for determining function calls to self, internal functions, and external functions. If not 
                passed, then these will not be used. Defaults to None.
            block (Union[CFGBasicBlock, int], optional): either a ``CFGBasicBlock`` or integer block_idx in a ``MemCFG``
                object. Used for determining function calls to self, internal functions, and external functions. If not 
                passed, then these will not be used. Defaults to None.
            newline_tup (Tuple[str, str], optional): the tuple to insert inbetween each passed string, or None to not 
                insert anything. Defaults to self.tokenizer.DEFAULT_NEWLINE_TUPLE
            match_instruction_address (bool, optional): if True, will match instruction addresses. If there is an immediate
                value at the start of a line (IE: start of a string in `strings`, or immediately after a Tokens.NEWLINE
                or Tokens.INSTRUCTION_START [ignoring any Tokens.SPACING]), then that token will be converted into a
                Tokens.INSTRUCTION_ADDRESS token. If there is a Tokens.COLON immediately after that token (again, ignoring
                any Tokens.SPACING), then that first Tokens.COLON match will be appended (along with any inbetween Tokens.SPACING)
                to that Tokens.INSTRUCTION_ADDRESS token. For example, using the x86 tokenization scheme:

                    - "0x1234: add rax rax" -> [(Tokens.INSTRUCTION_ADDRESS, '0x1234:'), ...]
                    - "  0x1234     : add rax rax" -> [(Tokens.SPACING, '  '), (Tokens.INSTRUCTION_ADDRESS, '0x1234     :'), ...]
                    - "0x1234 add rax rax" -> [(Tokens.INSTRUCTION_ADDRESS, '0x1234'), ...]
                
            kwargs (Any): extra kwargs to pass along to tokenization method, and to store in normalizer state

        Returns:
            List[str]: a list of normalized string instruction lines
        """
        return self.normalize(*strings, cfg=cfg, block=block, newline_tup=newline_tup, match_instruction_address=match_instruction_address)
    
    def __eq__(self, other):
        return type(self) == type(other) and all(t1 == t2 for t1, t2 in zip(self._arch_to_norm.items(), other.normalizers.items()))
    
    def __hash__(self):
        return hash_obj(self._arch_to_norm, return_int=True)
    
    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, ', '.join([repr(n) for n in self._arch_to_norm.values()]))
    
    def __str__(self):
        return '-'.join([self.__class__.__name__.lower()] + [str(n) for n in self._arch_to_norm.values()])



from .x86_tokenizer import X86_DEFAULT_TOKENS, X86_DEFAULT_TOKENIZER, X86BaseTokenizer
from .x86_normalizers import X86BaseNormalizer, X86DeepBinDiffNormalizer, X86SafeNormalizer, X86InnerEyeNormalizer, \
    X86DeepSemanticNormalizer, X86CompressedStatsNormalizer, X86HPCDataNormalizer
